import argparse
import json
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import geopandas as gpd
import shapely
from shapely.geometry import Point
import seaborn as sns
import matplotlib.pyplot as plt
from io import BytesIO
import base64
import scipy.stats as stats
from concurrent.futures import ProcessPoolExecutor
from tqdm import tqdm
import os
import shutil
import gzip
import cv2
import scanpy as sc
import scipy.io

def from_h5_to_files(fold):
    ad = sc.read_10x_h5(fold + "/filtered_feature_bc_matrix.h5")
    features = pd.DataFrame(ad.var.index)
    features[1] = features[0]
    features[2] = "Gene Expression"
    features.to_csv(os.path.join(fold, "features.tsv"), sep="\t", index=False, header=False)
    pd.DataFrame(ad.obs.index).to_csv(os.path.join(fold, "barcodes.tsv"), sep="\t", index=False, header=False)
    scipy.io.mmwrite(os.path.join(fold, "matrix.mtx"), ad.X.T)

def load_data(folder):
    dataset_name = folder.split("/")[-2] 
    print(f"Loading in {dataset_name}")
    
    barcodes_file = folder + "/barcodes.tsv"
    features_file = folder + "/features.tsv"
    matrix_file = folder + "/matrix.mtx"
    
    if not os.path.exists(barcodes_file) or not os.path.exists(features_file) or not os.path.exists(matrix_file):
        print(f"Missing files in {folder}. Attempting to generate from filtered_feature_bc_matrix.h5.")
        from_h5_to_files(folder)
    
    barcodes = pd.read_csv(barcodes_file, header=None, sep="\t")
    barcodes.columns = ["Barcode_ID"]
    barcodes["Barcode_ID"] = [f"{barcode}_{dataset_name}" for barcode in barcodes["Barcode_ID"]]
    barcodes["index_col"] = barcodes.index + 1
    
    tissue_positions_list = pd.read_csv(folder + "/spatial/tissue_positions_list.csv", header=None)
    tissue_positions_list.columns = ["barcode", "in_tissue", "tissue_col", "tissue_row", "x", "y"]
    tissue_positions_list["barcode"] = [f"{barcode}_{dataset_name}" for barcode in tissue_positions_list["barcode"]]
    tissue_positions_list["dataset"] = dataset_name
    barcode_indices = []
    for barcode in tissue_positions_list["barcode"]:
        try:
            barcode_indices.append(barcodes[barcodes["Barcode_ID"] == barcode].index_col.values[0])
        except:
            barcode_indices.append(-1)
            pass
    tissue_positions_list["barcode_id"] = barcode_indices
    
    features = pd.read_csv(features_file, header=None, sep="\t")
    if features.shape[1] == 3:
        features.columns = ["Gene_ID", "Gene_Name", "Type"]
    elif features.shape[1] == 1:
        features.columns = ["Gene_ID"]
    features["index_col"] = features.index + 1
    
    matrix = pd.read_csv(matrix_file, header=3, sep=" ")
    matrix.columns = ["Gene_ID", "Barcode_ID", "Counts"]
    
    image_exists = os.path.exists(folder + "/spatial/tissue_hires_image.png")
    if image_exists:
        image = plt.imread(folder + "/spatial/tissue_hires_image.png")
    else:
        image = np.zeros((int(tissue_positions_list["x"].max()), int(tissue_positions_list["y"].max())))
    
    scalefactors = json.load(open(folder + "/spatial/scalefactors_json.json"))
    
    data = {
        "matrix": matrix,
        "tissue_positions_list": tissue_positions_list, 
        "features": features,
        "barcodes": barcodes,
        "image": image,
        "scalefactors": scalefactors
    }
    return dataset_name, data



def merge_data(folders, pv_format=False):
    dataset_names = []
    nested_dict = {}
    features_df = pd.DataFrame()
    barcodes_df = pd.DataFrame()
    
    scalefactor_hires_vals=[]
    for folder in folders:
        scalefactors = json.load(open(folder+"/spatial/scalefactors_json.json"))
        scalefactor_hires_vals.append(scalefactors["tissue_hires_scalef"])
    
    min_scalefactor = min(scalefactor_hires_vals)
    resize_factors=[]
    for scalefactor_hires in scalefactor_hires_vals:
        resize_factors.append(min_scalefactor/scalefactor_hires)
    
    scalefactor_hires = min_scalefactor 
    min_scalefactor_folder = folders[scalefactor_hires_vals.index(min_scalefactor)]
    scalefactors_for_save = json.load(open(min_scalefactor_folder+"/spatial/scalefactors_json.json"))
    
    max_x_range = 0
    max_y_range = 0
    max_col_dims = 0
    max_row_dims = 0
    
    for folder,i in zip(folders,range(len(folders))):
        dataset_name, data = load_data(folder)
        dataset_names.append(dataset_name)
        nested_dict[dataset_name] = data

        #adjust image size
        image = data["image"]
        width = int(image.shape[1] * resize_factors[i])
        height = int(image.shape[0] * resize_factors[i])
        dim = (width, height)
        image = cv2.resize(image, dim, interpolation = cv2.INTER_AREA)
        nested_dict[dataset_name]["image"] = image
        
        barcodes_df_to_add = data["barcodes"].copy()
        barcodes_df_to_add["dataset"] = dataset_name
        barcodes_df = pd.concat([barcodes_df,barcodes_df_to_add])
        
        features_df_to_add = data["features"].copy()
        features_df_to_add["dataset"] = dataset_name
        features_df = pd.concat([features_df,features_df_to_add])
        
        if pv_format:
            pv_cell_hex = pd.read_csv(folder+"/spatial/pv_cell_hex.csv",header=None)
            pv_cell_hex.columns = ["Cell_ID","Hexagon_ID","Count"]
            pv_cell_hex["Cell_ID"] = [f"{cell_id}_{dataset_name}" for cell_id in pv_cell_hex["Cell_ID"]]
            nested_dict[dataset_name]["pv_cell_hex"] = pv_cell_hex
            arguments = json.load(open(folder+"arguments.json")) 
            nested_dict[dataset_name]["arguments"] = arguments
        
        tissue_positions_list = data["tissue_positions_list"]
        max_x_range = max(max_x_range,tissue_positions_list["x"].max())*1.1
        max_y_range = max(max_y_range,tissue_positions_list["y"].max())*1.1
        max_col_range = max(max_col_dims,tissue_positions_list["tissue_col"].max())*1.1
        max_row_range = max(max_row_dims,tissue_positions_list["tissue_row"].max())*1.1
        
    return dataset_names, nested_dict, features_df, barcodes_df, max_x_range, max_y_range, max_col_range, max_row_range, scalefactor_hires, scalefactors_for_save, resize_factors

def consensus_features(features_df):
    features_df_all = features_df.copy()
    unique_gene_ids = features_df_all["Gene_ID"].unique()
    features_df_all["consensus_index"] = features_df_all["Gene_ID"].apply(lambda x: np.where(unique_gene_ids==x)[0][0]+1)
    return features_df_all

def adjust_matrix_barcodes(nested_dict, features_df_all, pv_format=False):
    n_barcodes_before = 0
    for dataset_name in nested_dict.keys():
        print(dataset_name)
        matrix = nested_dict[dataset_name]["matrix"]
        features_all = features_df_all[features_df_all["dataset"]==dataset_name]
        for gene_id in matrix["Gene_ID"].unique():
            original_index,consensus_index = features_all[features_all["index_col"]==gene_id][["index_col","consensus_index"]].values[0]
            matrix.loc[matrix["Gene_ID"]==gene_id,"Gene_ID"] = consensus_index
        
        matrix["Barcode_ID"] = matrix["Barcode_ID"] + n_barcodes_before
        nested_dict[dataset_name]["matrix"] = matrix
        if pv_format:
            pv_cell_hex = nested_dict[dataset_name]["pv_cell_hex"]
            pv_cell_hex["Hexagon_ID"] = pv_cell_hex["Hexagon_ID"] + n_barcodes_before
            nested_dict[dataset_name]["pv_cell_hex"] = pv_cell_hex
            
        n_barcodes_before += len(nested_dict[dataset_name]["barcodes"])
    return nested_dict, n_barcodes_before

def stitch_images(nested_dict, grid, max_x_range, max_y_range, max_col_range, max_row_range, scalefactor_hires):
    new_tissue_positions_list = pd.DataFrame(columns=["barcode","in_tissue","tissue_col","tissue_row","x","y"])
    stitched_image = np.zeros((int(max_x_range*scalefactor_hires*grid[0]),int(max_y_range*scalefactor_hires*grid[1]),3))
    number_of_datasets = len(nested_dict.keys())
    for i in range(grid[0]):
        for j in range(grid[1]):

            dataset_idx = i*grid[1]+j
            if dataset_idx >= number_of_datasets:
                break
            print(dataset_idx)
            dataset_names = list(nested_dict.keys())
            tissue_positions_list_to_add = nested_dict[dataset_names[dataset_idx]]["tissue_positions_list"].copy()
            tissue_positions_list_to_add["x"] = tissue_positions_list_to_add["x"] + i*max_x_range
            tissue_positions_list_to_add["y"] = tissue_positions_list_to_add["y"] + j*max_y_range
            tissue_positions_list_to_add["tissue_col"] = tissue_positions_list_to_add["tissue_col"] + i*max_col_range
            tissue_positions_list_to_add["tissue_row"] = tissue_positions_list_to_add["tissue_row"] + j*max_row_range
            new_tissue_positions_list = pd.concat([new_tissue_positions_list,tissue_positions_list_to_add])
            new_tissue_positions_list.reset_index(drop=True,inplace=True)
            
            image_to_add = nested_dict[dataset_names[dataset_idx]]["image"]
            if len(image_to_add.shape) == 2:
                image_to_add = np.stack((image_to_add,)*3,axis=-1)
            image_to_add = np.pad(image_to_add,((0,max(0,int(max_x_range*scalefactor_hires)-image_to_add.shape[0])),(0,max(0,int(max_y_range*scalefactor_hires)-image_to_add.shape[1])),(0,0)))
            image_to_add = image_to_add[0:int(max_x_range*scalefactor_hires),0:int(max_y_range*scalefactor_hires)]
            image_to_add = np.pad(image_to_add,((0,int(max_x_range*scalefactor_hires)-image_to_add.shape[0]),(0,int(max_y_range*scalefactor_hires)-image_to_add.shape[1]),(0,0)))
            stitched_image[i*int(max_x_range*scalefactor_hires):(i+1)*int(max_x_range*scalefactor_hires),j*int(max_y_range*scalefactor_hires):(j+1)*int(max_y_range*scalefactor_hires),:] = image_to_add
    
    new_tissue_positions_list.drop(["dataset","barcode_id"],axis=1,inplace=True)
    return new_tissue_positions_list, stitched_image

def save_output(folderpath, scalefactors_for_save, barcodes_df_all, new_tissue_positions_list, pv_format, nested_dict, dataset_names, features_df_all, stitched_image):
    if not os.path.exists(folderpath):
        os.makedirs(folderpath + '/spatial/')
    
    with open(folderpath +'/spatial/scalefactors_json.json', 'w') as f:
        json.dump(scalefactors_for_save, f)
        
    barcodes_df_all.drop("index_col",axis=1,inplace=True)
    barcodes_df_all.to_csv(folderpath+'/barcodes.tsv',sep='\t',index=False,header=False)
    
    with open(folderpath +'/barcodes.tsv', 'rb') as f_in:
        with gzip.open(folderpath +'/barcodes.tsv.gz', 'wb') as f_out:
            f_out.writelines(f_in)
            
    new_tissue_positions_list.to_csv(folderpath +'/spatial/tissue_positions_list.csv',index=False,header=False)
    
    if pv_format:
        pv_cell_hex = pd.concat([nested_dict[dataset_name]["pv_cell_hex"] for dataset_name in dataset_names])
        pv_cell_hex.to_csv(folderpath +'/spatial/pv_cell_hex.csv',index=False,header=False)
        
    features_df_all = features_df_all.sort_values("consensus_index").drop_duplicates(subset=["Gene_ID"])
    features_df_all["Gene_Name"]=features_df_all["Gene_ID"]
    features_df_all["Type"]="Gene expression"
    features_df_all = features_df_all[["Gene_ID","Gene_Name","Type"]]
    features_df_all.to_csv(folderpath +'/features.tsv',sep='\t',index=False,header=False)
    
    with open(folderpath + '/features.tsv', 'rb') as f_in, gzip.open(folderpath + '/features.tsv.gz', 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)
        
    matrix_all = pd.concat([nested_dict[dataset_name]["matrix"] for dataset_name in dataset_names])
    metadata = f'%%MatrixMarket matrix coordinate integer general\n%metadata_json: {{"software_version": "Pseudovisium", "format_version": 1}}\n{len(features_df_all)} {barcodes_df_all.shape[0]} {len(matrix_all)}'
    matrix_all.to_csv(folderpath +'/matrix.mtx',index=False,header=False,sep=" ")
    with open(folderpath + '/matrix.mtx', 'r') as original: data = original.read()
    with open(folderpath + '/matrix.mtx', 'w') as modified: modified.write(metadata + '\n' + data)
    with open(folderpath + '/matrix.mtx', 'rb') as f_in, gzip.open(folderpath + '/matrix.mtx.gz', 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)
        
    plt.imsave(folderpath +'/spatial/tissue_hires_image.png', stitched_image)
    lowres_image = cv2.resize(stitched_image,(int(stitched_image.shape[1]/10),int(stitched_image.shape[0]/10)))
    plt.imsave(folderpath +'/spatial/tissue_lowres_image.png', lowres_image)

def merge_visium(folders, output_path, project_name, pv_format=False):
    dataset_names, nested_dict, features_df, barcodes_df, max_x_range, max_y_range, max_col_range, max_row_range, scalefactor_hires, scalefactors_for_save, resize_factors = merge_data(folders, pv_format)
    n_rows = max(3,int(np.sqrt(len(folders))))
    n_cols = int(np.ceil(len(folders)/n_rows))
    grid = [n_rows, n_cols]
    features_df_all = consensus_features(features_df)
    nested_dict, n_barcodes_before = adjust_matrix_barcodes(nested_dict, features_df_all, pv_format)
    barcodes_df_all = barcodes_df.copy()
    barcodes_df_all.reset_index(drop=True,inplace=True)
    barcodes_df_all["index_col"] = barcodes_df_all.index + 1
    barcodes_df_all.drop("dataset",axis=1,inplace=True)
    new_tissue_positions_list, stitched_image = stitch_images(nested_dict, grid, max_x_range, max_y_range,  max_col_range, max_row_range,scalefactor_hires)
    folderpath = output_path+ '/pseudovisium/' + project_name
    if os.path.exists(folderpath):
        shutil.rmtree(folderpath)
    save_output(folderpath, scalefactors_for_save, barcodes_df_all, new_tissue_positions_list, pv_format, nested_dict, dataset_names, features_df_all, stitched_image)

def main():
    parser = argparse.ArgumentParser(description="Merge Pseudovisium/Visium format files.")
    parser.add_argument("--folders", "-f", nargs="+", help="List of folders containing Pseudovisium/Visium output", required=True)
    parser.add_argument("--output_path", "-o", default="/Users/k23030440/", help="Output folder path")
    parser.add_argument("--project_name", "-p", default="visium_merged", help="Project name for output")
    parser.add_argument("--pv_format", action="store_true", help="Indicate if input is in Pseudovisium format")
    args = parser.parse_args()

    folders = [folder+"/" if not folder.endswith("/") else folder for folder in args.folders]
    merge_visium(folders, args.output_path, args.project_name, args.pv_format)

if __name__ == "__main__":
    main()



import argparse
import json
import matplotlib.pyplot as plt
from matplotlib.patches import RegularPolygon
import numpy as np
import mpld3
import pandas as pd
from pysal.lib import weights
from pysal.explore import esda
import geopandas as gpd
import shapely
from shapely.geometry import Point
import seaborn as sns
import matplotlib.pyplot as plt
from io import BytesIO
import base64
import scipy.stats as stats
from concurrent.futures import ProcessPoolExecutor
from tqdm import tqdm
import os
import datetime
import matplotlib.pyplot as plt
from matplotlib.collections import PathCollection
from adjustText import adjust_text
from io import BytesIO
import base64


def generate_qc_report(folders, output_folder="/Users/k23030440/", gene_names=["RYR3", "AQP4", "THBS1"], include_morans_i=False,max_workers=4):
    #if any entry in folders lacks final /, then add
    folders = [folder if folder[-1]=="/" else folder + "/" for folder in folders]
    #same with output_folder
    output_folder = output_folder if output_folder[-1]=="/" else output_folder + "/"

    replicates_data = []
    for folder in folders:
        # Extract the dataset name from the folder path
        dataset_name = folder.split("/")[-2]

        # Load files
        tissue_positions_list = pd.read_csv(folder + "spatial/tissue_positions_list.csv", header=None)
        tissue_positions_list.columns = ["barcode", "in_tissue", "tissue_col", "tissue_row", "x", "y"]
        matrix = pd.read_csv(folder + "matrix.mtx", header=3, sep=" ")
        matrix.columns = ["Gene_ID", "Barcode_ID", "Counts"]
        features = pd.read_csv(folder + "features.tsv", header=None, sep="\t")
        features.columns = ["Gene_ID", "Gene_Name", "Type"]
        features["index_col"] = features.index + 1
        barcodes = pd.read_csv(folder + "barcodes.tsv", header=None, sep="\t")
        barcodes.columns = ["Barcode_ID"]
        barcodes["index_col"] = barcodes.index + 1

        arguments = json.load(open(folder + "arguments.json"))
        #if cell_id_colname is not NA, then we have cell info
        cell_info= True if arguments["cell_id_colname"]!="NA" else False
        quality_per_hexagon = arguments["quality_per_hexagon"]
        quality_per_probe = arguments["quality_per_probe"]

        if quality_per_hexagon:
            hexagon_quality = pd.read_csv(folder + "spatial/quality_per_hexagon.csv", header=None)
            #name cols as barcode, quality, count
            hexagon_quality.columns = ["Hexagon_ID", "Quality", "Count"]
            #add dataset name
            hexagon_quality["Dataset"] = dataset_name
            hexagons_above_100 = hexagon_quality[hexagon_quality["Count"]>100]
            hexagons_q_below_20 = hexagons_above_100[hexagons_above_100["Quality"]<20]
            pct_hexagons_q_below_20 = len(hexagons_q_below_20)/len(hexagons_above_100)
            

        if quality_per_probe:
            probe_quality = pd.read_csv(folder + "spatial/quality_per_probe.csv", header=None)
            #name cols as barcode, quality, count
            probe_quality.columns = ["Probe_ID", "Quality", "Count"]
            probe_quality["Dataset"] = dataset_name
            non_ctrl_probes = probe_quality[~probe_quality["Probe_ID"].str.contains("control|ctrl|pos|neg|Ctrl|blank|Control|Blank|BLANK")]
            non_ctrl_probes_q_below_20 = non_ctrl_probes[non_ctrl_probes["Quality"]<20]
            pct_non_ctrl_probes_q_below_20 = len(non_ctrl_probes_q_below_20)/len(non_ctrl_probes)
            
        
        if cell_info:
            pv_cell_hex = pd.read_csv(folder + "spatial/pv_cell_hex.csv", header=None)
            pv_cell_hex.columns = ["Cell_ID", "Hexagon_ID", "Count"]
            pv_cell_hex["Hexagon_ID"] = pv_cell_hex["Hexagon_ID"] + 1




        arguments = json.load(open(folder + "arguments.json"))
        hexagon_size = arguments["hexagon_size"]
        image_pixels_per_um = arguments["image_pixels_per_um"]

        tissue_positions_list = tissue_positions_list[tissue_positions_list["in_tissue"] == 1]
        tissue_positions_list["barcode_id"] = [barcodes[barcodes["Barcode_ID"] == barcode].index_col.values[0] for barcode in tissue_positions_list["barcode"]]
        in_tissue_barcodes = tissue_positions_list["barcode"].index + 1

        n_barcodes_per_tissue = len(tissue_positions_list)

        # Join the x and y columns from tissue_positions_list based on Barcode_ID and barcode_id
        matrix_joined = pd.merge(matrix, tissue_positions_list[["barcode_id", "x", "y"]], left_on="Barcode_ID", right_on="barcode_id")
        matrix_joined = pd.merge(matrix_joined, features[["Gene_ID", "index_col"]], left_on="Gene_ID", right_on="index_col")

        if quality_per_hexagon:
            #in hexagon_quality keep only rows where hexagon_id is in index_col
            hexagon_quality = hexagon_quality[hexagon_quality["Hexagon_ID"].isin(matrix_joined["barcode_id"])]
            matrix_joined = pd.merge(matrix_joined, hexagon_quality, left_on="barcode_id", right_on="Hexagon_ID")

        # Calculate key metrics
        grouped_matrix = matrix_joined.groupby("Barcode_ID")["Counts"].sum()
        number_of_hex_above_100 = np.sum(grouped_matrix > 50)

        grouped_matrix = matrix_joined.groupby("Gene_ID_y")["Barcode_ID"].count()
        pct5_plex = np.sum(grouped_matrix > 0.05 * n_barcodes_per_tissue)
        probe_names = grouped_matrix[grouped_matrix > 0.05 * n_barcodes_per_tissue].index.values

        grouped_matrix = matrix_joined.groupby("Barcode_ID")["Counts"].sum()
        median_counts = np.median(grouped_matrix)
        cv_counts = np.std(grouped_matrix) / np.mean(grouped_matrix)

        grouped_matrix = matrix_joined.groupby("Barcode_ID")["Gene_ID_y"].count()
        median_features = np.median(grouped_matrix)
        cv_features = np.std(grouped_matrix) / np.mean(grouped_matrix)

        number_of_genes = len(features[~features["Gene_ID"].str.contains("control|ctrl|pos|neg|Ctrl|blank|Control|Blank|BLANK")])

        neg_control_probes = features[features["Gene_ID"].str.contains("Probe")].index + 1
        neg_control_counts = np.sum(matrix[matrix["Gene_ID"].isin(neg_control_probes)]["Counts"])
        total_counts = np.sum(matrix["Counts"])
        prop_neg_control = neg_control_counts / total_counts
        if cell_info:
            pv_cell_hex_assigned = pv_cell_hex[(pv_cell_hex["Cell_ID"] != "UNASSIGNED") & (pv_cell_hex["Cell_ID"] != -1)]
            grouped_pv_cell_hex_assigned = pv_cell_hex_assigned.groupby("Hexagon_ID")["Cell_ID"].count()
            median_cells_per_hex = np.median(grouped_pv_cell_hex_assigned)

            counts_per_cell = pv_cell_hex_assigned.groupby("Cell_ID")["Count"].sum()
            median_counts_per_cell = np.median(counts_per_cell)

            pv_cell_hex_unassigned = pv_cell_hex[(pv_cell_hex["Cell_ID"] != "UNASSIGNED") | (pv_cell_hex["Cell_ID"] != -1)]
            #merge with grouped_pv_cell_hex_assigned
            grouped_pv_cell_hex_assigned_sum = pv_cell_hex_assigned.groupby("Hexagon_ID")["Count"].sum()
            merged_assigned_unassigned = pd.merge(grouped_pv_cell_hex_assigned_sum,pv_cell_hex_unassigned,left_on="Hexagon_ID",right_on="Hexagon_ID",how="inner")
            merged_assigned_unassigned = merged_assigned_unassigned.rename(columns={"Count_x":"Assigned_count","Count_y":"Unassigned_count"})
            merged_assigned_unassigned = merged_assigned_unassigned.fillna(0)
            #add pct_unassigned
            merged_assigned_unassigned["pct_unassigned"] = merged_assigned_unassigned["Unassigned_count"]/(merged_assigned_unassigned["Assigned_count"]+merged_assigned_unassigned["Unassigned_count"])
            #merge with matrix
            merged_assigned_unassigned = pd.merge(merged_assigned_unassigned,tissue_positions_list,left_on="Hexagon_ID",right_on="barcode_id",how="inner",suffixes=('_a', '_b'))
            median_unassigned_pct = np.median(merged_assigned_unassigned["pct_unassigned"])


            # merge tissue_positions_list with pv_cell_hex
            tissue_positions_pv_cell_hex = pd.merge(tissue_positions_list, pv_cell_hex, left_on="barcode_id", right_on="Hexagon_ID", how="inner")
            #sum up the number of Cell_ID in every Hexagon_ID but keep x and y intact
            tissue_positions_pv_cell_hex_sum = tissue_positions_pv_cell_hex.groupby(["x","y"]).agg({"Cell_ID":"count","Hexagon_ID":"first"}).reset_index()
            #remove multi level index
            tissue_positions_pv_cell_hex_sum.reset_index(inplace=True)
            #rename Cell_ID to count
            tissue_positions_pv_cell_hex_sum.rename(columns={"Cell_ID":"counts"},inplace=True)
            density_cv = np.std(tissue_positions_pv_cell_hex_sum["counts"])/np.mean(tissue_positions_pv_cell_hex_sum["counts"])
            density_morans_i = get_morans_i("density", tissue_positions_pv_cell_hex_sum, tissue_positions_list,max_workers=max_workers) 
            




        plot_df = not_working_probe_based_on_sum(matrix_joined, sample_id=dataset_name)
        not_working_probes = plot_df[plot_df["Probe category"]=="Bad"].index.values
        n_probes_not_working = len(not_working_probes)
        
        

        replicate_data = {
            "dataset_name": dataset_name,
            "metrics_table_data": {
                "Number of hexagons with at least 100 counts": int(number_of_hex_above_100),
                "Number of genes in at least 5% of hexagons": int(pct5_plex),
                "Median counts per hexagon": int(median_counts),
                "Median features per hexagon": int(median_features),
                "Number of genes": int(number_of_genes),
                "Proportion of neg_control probes": np.round(prop_neg_control, 3),
                "Number of bad probes (Sum)": n_probes_not_working,
                "Features CV": np.round(cv_features, 3),
                "Counts CV": np.round(cv_counts, 3),
                "Features Morans I": np.round(get_morans_i("features", matrix_joined, tissue_positions_list,max_workers=max_workers),3),
                "Counts Morans I": np.round(get_morans_i("counts", matrix_joined, tissue_positions_list,max_workers=max_workers),3)
            },
            "matrix_joined": matrix_joined,
            "features": features,
            "tissue_positions_list": tissue_positions_list,
            "hexagon_size": hexagon_size,
            "image_pixels_per_um": image_pixels_per_um,
            "barcodes": barcodes,
            "probe_sum_stripplot_df":plot_df
        }

        if include_morans_i:
            replicate_data["morans_i"] = get_morans_i("all", matrix_joined, tissue_positions_list,max_workers=max_workers)
            plot_df_morans_i = not_working_probe_based_on_morans_i(replicate_data["morans_i"], sample_id=dataset_name)
            not_working_probes = plot_df_morans_i[plot_df_morans_i["Probe category"]=="Bad"].index.values
            n_probes_not_working = len(not_working_probes)
            #add to replicate_data metrics_table_data
            replicate_data["metrics_table_data"]["Number of bad probes (Morans I)"] = n_probes_not_working
            replicate_data["morans_i_stripplot_df"] = plot_df_morans_i

        if cell_info:
            replicate_data["metrics_table_data"]["Median cells per hexagon"] = int(median_cells_per_hex)
            replicate_data["metrics_table_data"]["Median counts per cell"] = int(median_counts_per_cell)
            replicate_data["metrics_table_data"]["Median pct unassigned"] = np.round(median_unassigned_pct, 3)
            replicate_data["merged_assigned_unassigned"] = merged_assigned_unassigned
            replicate_data["cell_density_df"] = tissue_positions_pv_cell_hex_sum

            replicate_data["metrics_table_data"]["Density CV"] = np.round(density_cv, 3)
            replicate_data["metrics_table_data"]["Density Morans I"] = np.round(density_morans_i, 3)

        if quality_per_hexagon:
            replicate_data["metrics_table_data"]["Pct hexagons with quality below 20"] = np.round(pct_hexagons_q_below_20, 3)
            replicate_data["hexagon_quality"] = hexagon_quality
            replicate_data["metrics_table_data"]["Quality Morans I"] = np.round(get_morans_i("Quality", matrix_joined, tissue_positions_list,max_workers=max_workers),3)


        if quality_per_probe:
            replicate_data["metrics_table_data"]["Pct non-ctrl probes with quality below 20"] = np.round(pct_non_ctrl_probes_q_below_20, 3)
            replicate_data["probe_quality"] = probe_quality
            plot_df_quality_per_probe = not_working_probe_based_on_quality(probe_quality, sample_id=dataset_name)
            replicate_data["probe_quality_stripplot_df"] = plot_df_quality_per_probe

        replicates_data.append(replicate_data)

    html_code = generate_dashboard_html(replicates_data, gene_names, include_morans_i,quality_per_hexagon,quality_per_probe,cell_info)

    # Save HTML code to a file
    with open(output_folder + "metrics_dashboard.html", "w", encoding="utf-8") as html_file:
        html_file.write(html_code)
        print("HTML file generated successfully!")
    
    #in the same output folder generate a folder called pv_qc_ date
    data_output_folder = output_folder + "pv_qc_" + str(datetime.datetime.now().date())
    #if that dir exists_ add a number to the end
    i=1
    while os.path.exists(data_output_folder):
        data_output_folder = output_folder + "pv_qc_" + str(datetime.datetime.now().date()) + "_" + str(i)
        i+=1
    os.mkdir(data_output_folder)

    #save the metrics table as a csv
    metrics_table = pd.DataFrame([replicate_data["metrics_table_data"] for replicate_data in replicates_data])
    #add a column for the dataset name
    metrics_table["Dataset"] = [replicate_data["dataset_name"] for replicate_data in replicates_data]
    metrics_table.to_csv(data_output_folder + "/metrics_table.csv", index=False)

    #save probe_quality as a csv
    if quality_per_probe:
        probe_quality = pd.concat([replicate_data["probe_quality_stripplot_df"] for replicate_data in replicates_data])
        probe_quality.to_csv(data_output_folder + "/probe_quality.csv", index=False)

    #save morans i values for each gene
    if include_morans_i:
        morans_i = pd.concat([replicate_data["morans_i_stripplot_df"] for replicate_data in replicates_data])
        morans_i.to_csv(data_output_folder + "/morans_i.csv", index=False)

    #save sum of probes stripplot
    sum_stripplot = pd.concat([replicate_data["probe_sum_stripplot_df"] for replicate_data in replicates_data])
    sum_stripplot.to_csv(data_output_folder + "/sum_stripplot.csv", index=False)



def generate_dashboard_html(replicates_data, gene_names, include_morans_i,quality_per_hexagon,quality_per_probe,cell_info):
    metrics_html = """
        <div id="metric-details">
            <table>
                <thead>
                    <tr>
                        <th>Metric</th>
    """
    for replicate_data in replicates_data:
        metrics_html += f"""
                        <th>{replicate_data['dataset_name']}</th>
        """
    metrics_html += """
                    </tr>
                </thead>
                <tbody>
    """
    for metric in replicates_data[0]['metrics_table_data']:
        metrics_html += f"""
                    <tr>
                        <td>{metric}</td>
        """
        for replicate_data in replicates_data:
            metrics_html += f"""
                        <td>{replicate_data['metrics_table_data'][metric]}</td>
            """
        metrics_html += """
                    </tr>
        """
    metrics_html += """
                </tbody>
            </table>
        </div>
    """

    hexagon_plots_html = ""
    for i, gene_name in enumerate(gene_names):
        if i % 3 == 0:
            hexagon_plots_html += f"""
            <div class="row">
            """
        gene_found = False
        for replicate_data in replicates_data:
            if gene_name in replicate_data['features']['Gene_Name'].tolist():
                gene_found = True
                hexagon_df = get_df_for_gene(replicate_data['matrix_joined'], replicate_data['tissue_positions_list'], gene_name)
                hexagon_html = hexagon_plot_to_html(hexagon_df, replicate_data['hexagon_size'], replicate_data['image_pixels_per_um'], gene_name, replicate_data['dataset_name'])
                hexagon_plots_html += f"""
                <div class="col">
                    {hexagon_html}
                </div>
                """
            else:
                hexagon_plots_html += """
                <div class="col">
                    <p>Gene not found in dataset.</p>
                </div>
                """
        if gene_found:
            hexagon_plots_html += """
            <div class="col">
                <div id="colorbar"></div>
            </div>
            """
        if (i + 1) % 3 == 0 or i == len(gene_names) - 1:
            hexagon_plots_html += """
            </div>
            """

    # Section for nFeature Hexagon plots
    nfeature_hexagon_plots_html = ""
    for i, replicate_data in enumerate(replicates_data):
        if i % 3 == 0:
            nfeature_hexagon_plots_html += f"""
            <div class="row">
            """
        unique_features_per_hexagon = get_unique_features_per_hexagon(replicate_data['matrix_joined'])
        hexagon_html = hexagon_plot_to_html(unique_features_per_hexagon, replicate_data['hexagon_size'], replicate_data['image_pixels_per_um'], "nFeature", replicate_data['dataset_name'],replicate_data['metrics_table_data']["Features Morans I"])
        nfeature_hexagon_plots_html += f"""
        <div class="col">
            <h3>{replicate_data['dataset_name']}</h3>
            {hexagon_html}
        </div>
        """
        if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
            nfeature_hexagon_plots_html += """
            </div>
            """

    if cell_info:
        cell_density_hexagon_plots_html = ""
        for i, replicate_data in enumerate(replicates_data):
            if i % 3 == 0:
                cell_density_hexagon_plots_html += f"""
                <div class="row">
                """
            cell_density_df= replicate_data["cell_density_df"]
            hexagon_html = hexagon_plot_to_html(cell_density_df, replicate_data['hexagon_size'], replicate_data['image_pixels_per_um'], "Density", replicate_data['dataset_name'],replicate_data['metrics_table_data']["Density Morans I"])
            cell_density_hexagon_plots_html += f"""
            <div class="col">
                <h3>{replicate_data['dataset_name']}</h3>
                {hexagon_html}
            </div>
            """
            if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
                cell_density_hexagon_plots_html += """
                </div>
                """




    if quality_per_hexagon:
        quality_hexagon_plots_html = ""
        for i, replicate_data in enumerate(replicates_data):
            if i % 3 == 0:
                quality_hexagon_plots_html += f"""
                <div class="row">
                """
            hexagon_quality = get_quality_per_hexagon(replicate_data['matrix_joined'])
            hexagon_html = hexagon_plot_to_html(hexagon_quality, replicate_data['hexagon_size'], replicate_data['image_pixels_per_um'], "Quality", replicate_data['dataset_name'],replicate_data['metrics_table_data']["Quality Morans I"])
            quality_hexagon_plots_html += f"""
            <div class="col">
                <h3>{replicate_data['dataset_name']}</h3>
                {hexagon_html}
            </div>
            """
            if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
                quality_hexagon_plots_html += """
                </div>
                """

                

    total_hexagon_plots_html = ""
    for i, replicate_data in enumerate(replicates_data):
        if i % 3 == 0:
            total_hexagon_plots_html += f"""
            <div class="row">
            """
        total_counts_per_hexagon = get_total_counts_per_hexagon(replicate_data['matrix_joined'])
        hexagon_html = hexagon_plot_to_html(total_counts_per_hexagon, replicate_data['hexagon_size'], replicate_data['image_pixels_per_um'], "Total exp", replicate_data['dataset_name'],replicate_data['metrics_table_data']["Counts Morans I"])
        total_hexagon_plots_html += f"""
        <div class="col">
            <h3>{replicate_data['dataset_name']}</h3>
            {hexagon_html}
        </div>
        """
        if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
            total_hexagon_plots_html += """
            </div>
            """

    sums_comparison_html = ""
    pairs = [(i, j) for i in range(len(replicates_data)) for j in range(i + 1, len(replicates_data))]
    for i, (index1, index2) in enumerate(pairs):
        if i % 3 == 0:
            sums_comparison_html += f"""
            <div class="row">
            """
        sums1 = get_probe_sums(replicates_data[index1]['matrix_joined'])
        sums2 = get_probe_sums(replicates_data[index2]['matrix_joined'])
        sums_plot_html = plot_sums_to_html(sums1, sums2, replicates_data[index1]['dataset_name'], replicates_data[index2]['dataset_name'])
        sums_comparison_html += f"""
        <div class="col">
            <h3>{replicates_data[index1]['dataset_name']} vs {replicates_data[index2]['dataset_name']}</h3>
            {sums_plot_html}
        </div>
        """
        if (i + 1) % 3 == 0 or i == len(pairs) - 1:
            sums_comparison_html += """
            </div>
            """

    abundance_correlation_heatmap_html = plot_abundance_correlation_heatmap(replicates_data)

    morans_i_comparison_html = ""
    morans_i_heatmap_html = ""
    morans_i_stripplot_html = ""
    if include_morans_i:
        pairs = [(i, j) for i in range(len(replicates_data)) for j in range(i + 1, len(replicates_data))]
        for i, (index1, index2) in enumerate(pairs):
            if i % 3 == 0:
                morans_i_comparison_html += f"""
                <div class="row">
                """
            morans_i1 = replicates_data[index1]['morans_i']
            morans_i2 = replicates_data[index2]['morans_i']
            morans_i_plot_html = plot_morans_i_to_html(morans_i1, morans_i2, replicates_data[index1]['dataset_name'], replicates_data[index2]['dataset_name'])
            morans_i_comparison_html += f"""
            <div class="col">
                <h3>{replicates_data[index1]['dataset_name']} vs {replicates_data[index2]['dataset_name']}</h3>
                {morans_i_plot_html}
            </div>
            """
            if (i + 1) % 6 == 0 or i == len(pairs) - 1:
                morans_i_comparison_html += """
                </div>
                """
        morans_i_heatmap_html = plot_morans_i_correlation_heatmap(replicates_data)
        for i, replicate_data in enumerate(replicates_data):
            if i % 3 == 0:
                morans_i_stripplot_html += f"""
                <div class="row">
                """
            morans_i_stripplot_df = replicate_data['morans_i_stripplot_df']
            morans_i_stripplot_html += f"""
            <div class="col">
                <h3>{replicate_data['dataset_name']}</h3>
                {probe_stripplot(morans_i_stripplot_df, sample_id=replicate_data['dataset_name'],legend=True, col_to_plot="Morans_I")}
            </div>
            """
            if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
                morans_i_stripplot_html += """
                </div>
                """

    sums_i_stripplot_html = ""
    for i, replicate_data in enumerate(replicates_data):
        if i % 3 == 0:
            sums_i_stripplot_html += f"""
            <div class="row">
            """
        plot_df = replicate_data['probe_sum_stripplot_df']
        sums_i_stripplot_html += f"""
        <div class="col">
            <h3>{replicate_data['dataset_name']}</h3>
            {probe_stripplot(plot_df, sample_id=replicate_data['dataset_name'],legend=True, col_to_plot="log_counts")}
        </div>
        """
        if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
            sums_i_stripplot_html += """
            </div>
            """

    quality_stripplot_html = ""
    if quality_per_probe:
        for i, replicate_data in enumerate(replicates_data):
            if i % 3 == 0:
                quality_stripplot_html += f"""
                <div class="row">
                """
            plot_df = replicate_data['probe_quality_stripplot_df']
            quality_stripplot_html += f"""
            <div class="col">
                <h3>{replicate_data['dataset_name']}</h3>
                {probe_stripplot(plot_df, sample_id=replicate_data['dataset_name'],legend=True, col_to_plot="Quality")}
            </div>
            """
            if (i + 1) % 3 == 0 or i == len(replicates_data) - 1:
                quality_stripplot_html += """
                </div>
                """

    html_code = f"""
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Pseudovisium QC</title>
        <style>
            body {{
                font-family: Arial, sans-serif;
            }}
            table {{
                border-collapse: collapse;
                width: 100%;
            }}
            th, td {{
                padding: 8px;
                text-align: left;
                border-bottom: 1px solid #ddd;
            }}
            th {{
                background-color: #f2f2f2;
            }}
            .row {{
                display: flex;
                flex-wrap: wrap;
                margin: 0 -20px;
            }}
            .col {{
                flex: 0 0 22%;
                max-width: 22%;
                padding: 0 20px;
                margin: 0 10px;
            }}
            .plot-container {{
                display: none;
            }}
        </style>
    </head>
    <body>
        <h1 style="text-align:center;">Pseudovisium QC</h1>
        <p style="text-align:center;">Written by Bence Kover (2024)</p>
        <div class="container">
            <div class="dropdown">
                <label for="metrics-select">Select Metric:</label>
                <select id="metrics-select">
                    <option value="table">Table of Key Metrics</option>
                    <option value="plot">Hexagon Plots</option>
                    <option value="sums-comparison">Sums Comparison</option>
                    <option value="abundance-correlation-heatmap">Abundance Correlation Heatmap</option>
                    {'<option value="morans-i-heatmap">Morans I Heatmap</option>' if include_morans_i else ""}
                    {'<option value="morans-i-stripplot">Morans I Stripplot</option>' if include_morans_i else ""}
                    <option value="sums-i-stripplot">Sums I Stripplot</option>
                    <option value="nfeature_hexagon_plots">nFeature hexagon plots</option>
                    {'<option value="quality_hexagon_plots">Quality hexagon plots</option>' if quality_per_hexagon else ""}
                    <option value="total_hexagon_plots">Total hexagon plots</option>
                    {'<option value="morans-i-comparison">Pairwise Morans I Plots</option>' if include_morans_i else ""}
                    {'<option value="probe_quality_stripplot">Probe Quality Stripplot</option>' if quality_per_probe else ""}
                    {'<option value="cell_density_hexagon_plots">Cell Density Hexagon Plots</option>' if cell_info else ""}
                </select>
            </div>
            <div id="metric-details-container">
                {metrics_html}
            </div>
            <div id="plot" class="plot-container">
                {hexagon_plots_html}
            </div>
            <div id="sums-comparison" class="plot-container">
                {sums_comparison_html}
            </div>
            <div id="abundance-correlation-heatmap" class="plot-container">
                {abundance_correlation_heatmap_html}
            </div>
            <div id="morans-i-heatmap" class="plot-container">
                {morans_i_heatmap_html}
            </div>
            <div id="morans-i-stripplot" class="plot-container">
                {morans_i_stripplot_html}
            </div>
            <div id="sums-i-stripplot" class="plot-container">
                {sums_i_stripplot_html}
            </div>
            <div id="nfeature_hexagon_plots" class="plot-container">
                {nfeature_hexagon_plots_html}
            </div>
            <div id="quality_hexagon_plots" class="plot-container">
                {quality_hexagon_plots_html}
            </div>
            <div id="total_hexagon_plots" class="plot-container">
                {total_hexagon_plots_html}
            </div>
            <div id="morans-i-comparison" class="plot-container">
                {morans_i_comparison_html}
            </div>
            <div id="probe_quality_stripplot" class="plot-container">
                {quality_stripplot_html}
            </div>
            <div id="cell_density_hexagon_plots" class="plot-container">
                {cell_density_hexagon_plots_html}
            </div>
        </div>

        <script>
            const select = document.getElementById('metrics-select');
            const metricDetailsContainer = document.getElementById('metric-details-container');
            const plotContainer = document.getElementById('plot');
            const sumsComparisonContainer = document.getElementById('sums-comparison');
            const abundanceCorrelationHeatmapContainer = document.getElementById('abundance-correlation-heatmap');
            const moransIHeatmapContainer = document.getElementById('morans-i-heatmap');
            const moransIStripplotContainer = document.getElementById('morans-i-stripplot');
            const sumsIStripplotContainer = document.getElementById('sums-i-stripplot');
            const nfeatureHexagonContainer = document.getElementById('nfeature_hexagon_plots');
            const qualityHexagonContainer = document.getElementById('quality_hexagon_plots');
            const totalHexagonContainer = document.getElementById('total_hexagon_plots');
            const moransIComparisonContainer = document.getElementById('morans-i-comparison');
            const probeQualityStripplotContainer = document.getElementById('probe_quality_stripplot');
            const cellDensityHexagonContainer = document.getElementById('cell_density_hexagon_plots');


            function updateMetricDetails() {{
                const selectedMetric = select.value;
                const plotContainers = document.querySelectorAll('.plot-container');

                // Hide all plot containers
                plotContainers.forEach(container => {{
                    container.style.display = 'none';
                }});

                // Hide the metric details container
                metricDetailsContainer.style.display = 'none';

                if (selectedMetric === 'table') {{
                    metricDetailsContainer.style.display = 'block';
                }} else if (selectedMetric === 'plot') {{
                    plotContainer.style.display = 'block';
                }} else if (selectedMetric === 'sums-comparison') {{
                    sumsComparisonContainer.style.display = 'block';
                }} else if (selectedMetric === 'abundance-correlation-heatmap') {{
                    abundanceCorrelationHeatmapContainer.style.display = 'block';
                }} else if (selectedMetric === 'morans-i-heatmap') {{
                    moransIHeatmapContainer.style.display = 'block';
                }} else if (selectedMetric === 'morans-i-stripplot') {{
                    moransIStripplotContainer.style.display = 'block';
                }} else if (selectedMetric === 'sums-i-stripplot') {{
                    sumsIStripplotContainer.style.display = 'block';
                }} else if (selectedMetric === 'nfeature_hexagon_plots') {{
                    nfeatureHexagonContainer.style.display = 'block';
                }} else if (selectedMetric === 'quality_hexagon_plots') {{
                    qualityHexagonContainer.style.display = 'block';
                }} else if (selectedMetric === 'total_hexagon_plots') {{
                    totalHexagonContainer.style.display = 'block';
                }} else if (selectedMetric === 'morans-i-comparison') {{
                    moransIComparisonContainer.style.display = 'block';
                }}
                else if (selectedMetric === 'probe_quality_stripplot') {{
                    probeQualityStripplotContainer.style.display = 'block';
                }}
                else if (selectedMetric === 'cell_density_hexagon_plots') {{
                    cellDensityHexagonContainer.style.display = 'block';
                }}

            }}

            select.addEventListener('change', updateMetricDetails);

            // Initial update
            updateMetricDetails();
        </script>
    </body>
    </html>
    """
    return html_code


def not_working_probe_based_on_sum(matrix_joined,sample_id="Sample1"):
    grouped_matrix = matrix_joined.groupby("Gene_ID_y")["Counts"].sum()
    #where index has control|blank|Control|Blank|BLANK in it
    grouped_matrix_neg_probes = grouped_matrix[grouped_matrix.index.str.contains("control|blank|Control|Blank|BLANK")]
    grouped_matrix_true_probes = grouped_matrix[~grouped_matrix.index.str.contains("control|blank|Control|Blank|BLANK")]     

    #create a plot_df that is grouped_matrix and a column specifying whether the gene is a neg control or not
    plot_df = pd.DataFrame(grouped_matrix)
    plot_df["Probe category"] = [1 if gene in grouped_matrix_neg_probes.index.values else 0 for gene in plot_df.index.values]
    plot_df["gene"] = plot_df.index.values
    plot_df["log_counts"] = np.log10(plot_df["Counts"])

    #iterate through the true probes and see whether they are significantly outside of the distribution of the neg probes
    for gene in grouped_matrix_true_probes.index.values:
        gene_count = plot_df.loc[gene,"log_counts"]
        neg_probes_count = plot_df[plot_df["Probe category"]==1]["log_counts"]
        #get pval with ttest
        mean = np.mean(neg_probes_count)
        std = np.std(neg_probes_count)
        p_val = stats.norm.cdf(gene_count, loc=mean, scale=std)
        p_val = 1-p_val
        if p_val*len(grouped_matrix_true_probes)<0.05:
            plot_df.loc[gene,"Probe category"] = 2
        
    plot_df["Probe category"] = ["Neg_control" if x==1 else "Bad" if x==0 else "Good" for x in plot_df["Probe category"]]
    plot_df["Sample"] = sample_id
    #order based on probe category
    plot_df = plot_df.sort_values("Probe category")
    return plot_df



import numpy as np
import matplotlib.pyplot as plt
from adjustText import adjust_text
from io import BytesIO
import base64


def probe_stripplot(plot_df, col_to_plot="log_counts", sample_id="Sample 1", legend=False):
    fig, ax = plt.subplots(figsize=(2, 2.5))
    # Define jitter amount
    jitter = 0.1

    # Create a dictionary to store x and y values for each point
    points = {"x": [], "y": [], "cat": [], "index": []}

    # Assign x and y values with jitter for each point
    for index, row in plot_df.iterrows():
        cat = row["Probe category"]
        x = 0 + np.random.uniform(-jitter, jitter)
        y = row[col_to_plot]
        points["x"].append(x)
        points["y"].append(y)
        points["cat"].append(cat)
        points["index"].append(index)

    # Plot the points for each category
    for cat in ["Neg_control", "Good", "Bad"]:
        mask = [c == cat for c in points["cat"]]
        ax.scatter([x for x, m in zip(points["x"], mask) if m],
                   [y for y, m in zip(points["y"], mask) if m],
                   alpha=0.8, label=cat)

    # Draw a straight line at the mean of the neg controls
    neg_control_mean = np.mean(plot_df[plot_df["Probe category"] == "Neg_control"][col_to_plot])
    ax.axhline(neg_control_mean, color="red", linestyle="--")

    # Add labels for the top 10 lowest score Bad probes
    bad_probes = plot_df[plot_df["Probe category"] == "Bad"].sort_values(col_to_plot).head(10)
    colname = "gene" if "gene" in plot_df.columns else "Probe_ID"

    texts = []
    for index, row in bad_probes.iterrows():
        x = points["x"][points["index"].index(index)]
        y = points["y"][points["index"].index(index)]
        txt = row[colname]
        texts.append(ax.text(x, y, txt, fontsize=8, ha='left', va='center'))

    adjust_text(texts, arrowprops=dict(arrowstyle="-", color='black', lw=0.5))

    # Remove x-axis ticks and label
    ax.set_xticks([])
    ax.set_xlabel("")
    #add y label based on col_to_plot
    ax.set_ylabel(col_to_plot)

    # Add a legend if requested
    if legend:
        ax.legend(fontsize='small', loc='upper left', bbox_to_anchor=(1, 1))

    # Convert the plot to HTML
    img_buffer = BytesIO()
    fig.savefig(img_buffer, format='png', bbox_inches='tight')
    img_str = base64.b64encode(img_buffer.getvalue()).decode()
    html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

    plt.close(fig)
    return html_fig

def not_working_probe_based_on_quality(probe_quality, sample_id="Sample1"):
    probe_quality_neg_probes = probe_quality[probe_quality["Probe_ID"].str.contains("control|blank|Control|Blank|BLANK")]
    probe_quality_true_probes = probe_quality[~probe_quality["Probe_ID"].str.contains("control|blank|Control|Blank|BLANK")]
    plot_df = probe_quality.reset_index(drop=True)
    plot_df["Probe category"] = [1 if gene in probe_quality_neg_probes.Probe_ID.values else 0 for gene in plot_df.Probe_ID.values]

    #iterate through the true probes and see whether they are significantly outside of the distribution of the neg probes
    for gene in probe_quality_true_probes.Probe_ID.values:
        plot_df_gene_index = plot_df[plot_df["Probe_ID"]==gene].index[0]
        quality = plot_df[plot_df["Probe_ID"]==gene]["Quality"]
        neg_probes_quality = plot_df[plot_df["Probe category"]==1]["Quality"]
        mean = np.mean(neg_probes_quality)
        std = np.std(neg_probes_quality)
        p_val = stats.norm.cdf(quality, loc=mean, scale=std)
        p_val = 1-p_val
        if p_val*len(probe_quality_true_probes)<0.05:
            plot_df.loc[plot_df_gene_index,"Probe category"] = 2
        
    plot_df["Probe category"] = ["Neg_control" if x==1 else "Bad" if x==0 else "Good" for x in plot_df["Probe category"]]
    plot_df["Sample"] = sample_id
    plot_df = plot_df.sort_values("Probe category")
    return plot_df




def not_working_probe_based_on_morans_i(morans_table, sample_id="Sample1"):
    morans_table_neg_probes = morans_table[morans_table.gene.str.contains("control|blank|Control|Blank|BLANK")]
    morans_table_true_probes = morans_table[~morans_table.gene.str.contains("control|blank|Control|Blank|BLANK")]     

    #create a plot_df that is grouped_matrix and a column specifying whether the gene is a neg control or not
    plot_df = morans_table.reset_index(drop=True)
    plot_df["Probe category"] = [1 if gene in morans_table_neg_probes.gene.values else 0 for gene in plot_df.gene.values]

    #iterate through the true probes and see whether they are significantly outside of the distribution of the neg probes
    for gene in morans_table_true_probes.gene.values:
        plot_df_gene_index = plot_df[plot_df["gene"]==gene].index[0]
        morans_i_s = plot_df[plot_df["gene"]==gene]["Morans_I"]
        neg_probes_morans_i_s = plot_df[plot_df["Probe category"]==1]["Morans_I"]
        mean = np.mean(neg_probes_morans_i_s)
        std = np.std(neg_probes_morans_i_s)
        p_val = stats.norm.cdf(morans_i_s, loc=mean, scale=std)
        p_val = 1-p_val
        if p_val*len(morans_table_true_probes)<0.05:
            plot_df.loc[plot_df_gene_index,"Probe category"] = 2
    #save the name of those genes with 0
    plot_df["Probe category"] = ["Neg_control" if x==1 else "Bad" if x==0 else "Good" for x in plot_df["Probe category"]]
    plot_df["Sample"] = sample_id
    plot_df = plot_df.sort_values("Probe category")
    return plot_df



def get_unique_features_per_hexagon(matrix_joined):
    unique_features_per_hexagon = matrix_joined.groupby(['x', 'y'])['Gene_ID_y'].nunique().reset_index()
    unique_features_per_hexagon = unique_features_per_hexagon.rename(columns={"Gene_ID_y": "counts"})
    return unique_features_per_hexagon


def get_total_counts_per_hexagon(matrix_joined):
    total_counts_per_hexagon = matrix_joined.groupby(['x', 'y'])['Counts'].sum().reset_index()
    total_counts_per_hexagon = total_counts_per_hexagon.rename(columns={"Counts": "counts"})
    return total_counts_per_hexagon

def get_quality_per_hexagon(matrix_joined):
    hexagon_quality = matrix_joined.groupby("Barcode_ID").agg({"Quality":"first","x":"first","y":"first"})
    hexagon_quality = hexagon_quality.reset_index()
    #rename Quality to counts
    hexagon_quality = hexagon_quality.rename(columns={"Quality": "counts"})
    return hexagon_quality


from shapely.geometry import Point
import geopandas as gpd
from libpysal import weights
from esda import Moran
import numpy as np
import pandas as pd

# Functions used in generate_qc_report
def get_df_for_gene(matrix_joined, tissue_positions_list, gene_name, normalised=False):
    matrix_subset = matrix_joined[matrix_joined["Gene_ID_y"] == gene_name]
    matrix_subset.reset_index(drop=True, inplace=True)
    x = tissue_positions_list["x"]
    y = tissue_positions_list["y"]
    counts = np.zeros(len(tissue_positions_list))
    for i in range(len(matrix_subset)):
        count = matrix_subset["Counts"][i]
        barcode_id = matrix_subset["Barcode_ID"][i]
        x_, y_ = matrix_subset["x"][i], matrix_subset["y"][i]
        x = np.append(x, x_)
        y = np.append(y, y_)

        if normalised:
            count = count / np.sum(matrix_joined[matrix_joined["Barcode_ID"] == barcode_id]["Counts"])

        counts = np.append(counts, count)
    df = pd.DataFrame({"x": x, "y": y, "counts": counts})
    return df.sort_values("counts", ascending=False).drop_duplicates(subset=["x", "y"])



def hexagon_plot_to_html(hexagon_df, hexagon_size, image_pixels_per_um, gene_name, dataset_name,morans_i=None):
    fig, ax = plt.subplots(figsize=(3, 2.5))
    sc = ax.scatter(hexagon_df["x"], hexagon_df["y"], c=hexagon_df["counts"], cmap="viridis", s=2, alpha=0.6)
    for hx, hy in zip(hexagon_df["x"], hexagon_df["y"]):
        hexagon = RegularPolygon((hx, hy), numVertices=6, radius=hexagon_size * image_pixels_per_um / 0.865, alpha=0.2, edgecolor='k', orientation=np.pi / 2)
        ax.add_patch(hexagon)
    ax.set_xlim(0, max(hexagon_df["x"]))
    ax.set_ylim(0, max(hexagon_df["y"]))
    aspect = hexagon_df["x"].max() / hexagon_df["y"].max()
    ax.set_aspect(aspect)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(f"{gene_name} - {dataset_name}", loc='center', fontsize=8)

    # Add color bar
    sm = plt.cm.ScalarMappable(cmap="viridis", norm=plt.Normalize(vmin=hexagon_df["counts"].min(), vmax=hexagon_df["counts"].max()))
    sm._A = []
    cbar = plt.colorbar(sm, ax=ax, fraction=0.046, pad=0.04)
    cbar.ax.tick_params(labelsize=6)
    if morans_i:
        ax.set_title(f"{gene_name} - Morans I: {morans_i:.2f}", fontsize=8)

    # Convert plot to base64 encoded image string
    img_buffer = BytesIO()
    fig.savefig(img_buffer, format='png')
    img_str = base64.b64encode(img_buffer.getvalue()).decode()
    html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

    plt.close(fig)  # Close the plot to free memory

    return html_fig



def get_probe_sums(matrix_joined):
    sums = matrix_joined.groupby("Gene_ID_y").sum()
    sums["Counts"] = np.log10(sums["Counts"])
    sums = sums[["Counts"]]
    sums.reset_index(inplace=True)
    return sums


def plot_sums_to_html(sums1, sums2, dataset1_name, dataset2_name):
    common_probes = list(set(sums1["Gene_ID_y"]) & set(sums2["Gene_ID_y"]))

    if len(common_probes) < 20:
        fig, ax = plt.subplots(figsize=(4, 4))
        ax.text(0.5, 0.5, "Not enough overlapping probes", fontsize=12, ha='center')
        ax.axis('off')

        # Convert the plot to HTML
        img_buffer = BytesIO()
        fig.savefig(img_buffer, format='png')
        img_str = base64.b64encode(img_buffer.getvalue()).decode()
        html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

        plt.close(fig)  # Close the plot to free memory

        return html_fig
    else:
        # Filter to only contain common probes
        sums1 = sums1[sums1["Gene_ID_y"].isin(common_probes)]
        sums2 = sums2[sums2["Gene_ID_y"].isin(common_probes)]
        # Reorder
        sums1 = sums1.sort_values("Gene_ID_y")
        sums2 = sums2.sort_values("Gene_ID_y")

        sums_df = pd.merge(sums1, sums2, on="Gene_ID_y", suffixes=("_1", "_2"))
        fig, ax = plt.subplots(figsize=(4, 4))
        ax.scatter(sums_df["Counts_1"], sums_df["Counts_2"], alpha=0.3, s=20)
        ax.set_xlabel(f"Log10 Total expression {dataset1_name}", fontsize=8)
        ax.set_ylabel(f"Log10 Total expression {dataset2_name}", fontsize=8)
        ax.set_title(f"{dataset1_name} vs {dataset2_name}", fontsize=10)
        corr = sums_df["Counts_1"].corr(sums_df["Counts_2"])
        ax.text(np.quantile(sums_df["Counts_1"], 0.3), np.quantile(sums_df["Counts_2"], 0.95),
                f"Pearson correlation: {corr:.2f}", fontsize=8)

        sums_df["diff"] = np.abs(sums_df["Counts_1"] - sums_df["Counts_2"])
        sums_df = sums_df.sort_values("diff", ascending=False)

        texts = []
        for i in range(10):
            x = sums_df["Counts_1"].iloc[i]
            y = sums_df["Counts_2"].iloc[i]
            txt = sums_df["Gene_ID_y"].iloc[i]
            texts.append(ax.text(x, y, txt, fontsize=10))

        adjust_text(texts, arrowprops=dict(arrowstyle="-", color='black', lw=0.5))

        # Convert the plot to HTML
        img_buffer = BytesIO()
        fig.savefig(img_buffer, format='png', bbox_inches='tight')
        img_str = base64.b64encode(img_buffer.getvalue()).decode()
        html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

        plt.close(fig)  # Close the plot to free memory

        return html_fig



def plot_abundance_correlation_heatmap(replicates_data):
    # Calculate the sums of features for each replicate
    sums_data = []
    replicate_names = []
    for replicate_data in replicates_data:
        sums = get_probe_sums(replicate_data['matrix_joined'])
        sums_data.append(sums)
        replicate_names.append(replicate_data['dataset_name'])
    
    corr_matrix = pd.DataFrame(index=replicate_names, columns=replicate_names, dtype=float)
    min_common_probes = 20
    for i in range(len(sums_data)):
        for j in range(i + 1, len(sums_data)):
            common_probes = list(set(sums_data[i]["Gene_ID_y"]) & set(sums_data[j]["Gene_ID_y"]))
            if len(common_probes) >= min_common_probes:
                #filter dataframes only for common probes then reorder them to have the same order
                sums_data[i] = sums_data[i][sums_data[i]["Gene_ID_y"].isin(common_probes)]
                sums_data[j] = sums_data[j][sums_data[j]["Gene_ID_y"].isin(common_probes)]
                sums_data[i].sort_values("Gene_ID_y", inplace=True)
                sums_data[j].sort_values("Gene_ID_y", inplace=True)
                corr = sums_data[i]["Counts"].corr(sums_data[j]["Counts"])
                corr=round(corr,3)
                corr_matrix.iloc[i, j] = corr
                corr_matrix.iloc[j, i] = corr
            else:
                corr_matrix.iloc[i, j] = 0.0
                corr_matrix.iloc[j, i] = 0.0
            
    #make diagonal 1
    np.fill_diagonal(corr_matrix.values, 1)
    
    fig_dimension = len(replicates_data)

    clustermap = sns.clustermap(corr_matrix, cmap='coolwarm', annot=True, fmt='.2f', figsize=(fig_dimension, fig_dimension))

    # Convert the plot to HTML
    img_buffer = BytesIO()
    clustermap.savefig(img_buffer, format='png')
    img_str = base64.b64encode(img_buffer.getvalue()).decode()
    html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

    plt.close(clustermap.figure)
    return html_fig


def get_df_for_gene(matrix_joined, tissue_positions_list, gene_name, normalised=False):
    matrix_subset = matrix_joined[matrix_joined["Gene_ID_y"] == gene_name]
    matrix_subset.reset_index(drop=True, inplace=True)
    x = tissue_positions_list["x"]
    y = tissue_positions_list["y"]
    counts = np.zeros(len(tissue_positions_list))
    for i in range(len(matrix_subset)):
        count = matrix_subset["Counts"][i]
        barcode_id = matrix_subset["Barcode_ID"][i]
        x_, y_ = matrix_subset["x"][i], matrix_subset["y"][i]
        x = np.append(x, x_)
        y = np.append(y, y_)

        if normalised:
            count = count / np.sum(matrix_joined[matrix_joined["Barcode_ID"] == barcode_id]["Counts"])

        counts = np.append(counts, count)
    df = pd.DataFrame({"x": x, "y": y, "counts": counts})
    return df.sort_values("counts", ascending=False).drop_duplicates(subset=["x", "y"])


def process_gene(gene, matrix_joined, tissue_positions_list):
    gene_df = get_df_for_gene(matrix_joined, tissue_positions_list, gene, normalised=True)
    points = [Point(xy) for xy in zip(gene_df['x'], gene_df['y'])]
    gene_gdf = gpd.GeoDataFrame(gene_df, geometry=points)
    w = weights.KNN.from_dataframe(gene_gdf, k=6)
    w.transform = 'R'
    mi = Moran(gene_df["counts"], w, permutations=0)
    return {"gene": gene, "Morans_I": mi.I}

def get_morans_i(gene_name, matrix_joined, tissue_positions_list, max_workers=4):
    if gene_name == "all":
        unique_genes = matrix_joined["Gene_ID_y"].unique()
        #remove those unique genes which have less than total 100 counts
        lowly_expressed_genes = matrix_joined.groupby("Gene_ID_y")["Counts"].sum()
        lowly_expressed_genes = lowly_expressed_genes[lowly_expressed_genes<100].index.values
        unique_genes = [gene for gene in unique_genes if gene not in lowly_expressed_genes]
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            results = list(tqdm(executor.map(process_gene, unique_genes, [matrix_joined]*len(unique_genes), [tissue_positions_list]*len(unique_genes)), total=len(unique_genes), desc="Processing genes"))
        
        res_df = pd.DataFrame(results)
        #append the lowly expressed genes to the end of the dataframe with a morans i of 0
        res_df = pd.concat([res_df,pd.DataFrame({"gene":lowly_expressed_genes,"Morans_I":0})],axis=0)
        
        return res_df
    
    elif gene_name == "features":
        mat = matrix_joined.groupby("Barcode_ID").count()
        mat["x"] = mat.index.to_series().apply(lambda barcode: matrix_joined[matrix_joined["Barcode_ID"]==barcode]["x"].values[0])
        mat["y"] = mat.index.to_series().apply(lambda barcode: matrix_joined[matrix_joined["Barcode_ID"]==barcode]["y"].values[0])
        points = [Point(xy) for xy in zip(mat['x'], mat['y'])]
        gene_gdf = gpd.GeoDataFrame(mat, geometry=points)
        w = weights.KNN.from_dataframe(gene_gdf, k=6)
        w.transform = 'R'
        mi = esda.Moran(mat["Gene_ID_y"], w, permutations=0)
        return mi.I
    
    elif gene_name == "density":
        mat = matrix_joined.copy()
        points = [Point(xy) for xy in zip(mat['x'], mat['y'])]
        gene_gdf = gpd.GeoDataFrame(mat, geometry=points)
        w = weights.KNN.from_dataframe(gene_gdf, k=6)
        w.transform = 'R'
        mi = esda.Moran(mat["counts"], w, permutations=0)
        return mi.I
    
    elif gene_name == "counts":
        mat = matrix_joined.groupby("Barcode_ID").sum()
        mat["x"] = mat.index.to_series().apply(lambda barcode: matrix_joined[matrix_joined["Barcode_ID"]==barcode]["x"].values[0])
        mat["y"] = mat.index.to_series().apply(lambda barcode: matrix_joined[matrix_joined["Barcode_ID"]==barcode]["y"].values[0])
        points = [Point(xy) for xy in zip(mat['x'], mat['y'])]
        gene_gdf = gpd.GeoDataFrame(mat, geometry=points)
        w = weights.KNN.from_dataframe(gene_gdf, k=6)
        w.transform = 'R'
        mi = esda.Moran(mat["Counts"], w, permutations=0)
        return mi.I

    elif gene_name == "Quality":
        mat = matrix_joined.groupby("Barcode_ID").agg({"Quality":"mean"})
        mat["x"] = mat.index.to_series().apply(lambda barcode: matrix_joined[matrix_joined["Barcode_ID"]==barcode]["x"].values[0])
        mat["y"] = mat.index.to_series().apply(lambda barcode: matrix_joined[matrix_joined["Barcode_ID"]==barcode]["y"].values[0])
        points = [Point(xy) for xy in zip(mat['x'], mat['y'])]
        gene_gdf = gpd.GeoDataFrame(mat, geometry=points)
        w = weights.KNN.from_dataframe(gene_gdf, k=6)
        w.transform = 'R'
        mi = esda.Moran(mat["Quality"], w, permutations=0)
        return mi.I
    
    else:
        gene_df = get_df_for_gene(matrix_joined, tissue_positions_list, gene_name, normalised=True)
        points = [Point(xy) for xy in zip(gene_df['x'], gene_df['y'])]
        gene_gdf = gpd.GeoDataFrame(gene_df, geometry=points)
        w = weights.KNN.from_dataframe(gene_gdf, k=6)
        w.transform = 'R'
        mi = esda.Moran(gene_df["counts"], w, permutations=0)
        return mi.I



def plot_morans_i_to_html(morans_i1, morans_i2, dataset1_name, dataset2_name):
    common_probes = list(set(morans_i1["gene"]) & set(morans_i2["gene"]))
    #print length of common_probes "in scatter plot"
    print("In scatterplot, number of common probes: ",len(common_probes))
    if len(common_probes) < 20:
        fig, ax = plt.subplots(figsize=(4, 4))
        ax.text(0.5, 0.5, "Not enough overlapping probes", fontsize=12, ha='center')
        ax.axis('off')

        # Convert the plot to HTML
        img_buffer = BytesIO()
        fig.savefig(img_buffer, format='png')
        img_str = base64.b64encode(img_buffer.getvalue()).decode()
        html_fig = f'<img src="data:image/png;base64,{img_str}"/>'
        
        plt.close(fig)  # Close the plot to free memory

        return html_fig
    else:
        morans_i_df = pd.merge(morans_i1, morans_i2, on="gene", suffixes=("_1", "_2"))
        morans_i_df = morans_i_df[morans_i_df["gene"].isin(common_probes)]
        fig, ax = plt.subplots(figsize=(4, 4))
        ax.scatter(morans_i_df["Morans_I_1"], morans_i_df["Morans_I_2"], alpha=0.3, s=20)
        ax.set_xlabel(f"Moran's I {dataset1_name}", fontsize=8)
        ax.set_ylabel(f"Moran's I {dataset2_name}", fontsize=8)
        ax.set_title(f"{dataset1_name} vs {dataset2_name}", fontsize=10)
        corr = morans_i_df["Morans_I_1"].corr(morans_i_df["Morans_I_2"])
        ax.text(0.5, 0.1, f"Pearson correlation: {corr:.2f}", fontsize=8, ha='center')
        morans_i_df["diff"] = np.abs(morans_i_df["Morans_I_1"] - morans_i_df["Morans_I_2"])
        morans_i_df = morans_i_df.sort_values("diff", ascending=False)
        for i in range(10):
            ax.text(morans_i_df["Morans_I_1"].iloc[i], morans_i_df["Morans_I_2"].iloc[i], morans_i_df["gene"].iloc[i], fontsize=10)

        # Convert the plot to HTML
        img_buffer = BytesIO()
        fig.savefig(img_buffer, format='png')
        img_str = base64.b64encode(img_buffer.getvalue()).decode()
        html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

        plt.close(fig)  # Close the plot to free memory

        return html_fig


def plot_morans_i_correlation_heatmap(replicates_data):
    #get the morans i for each dataset
    morans_i_data = []
    replicate_names = []
    for replicate_data in replicates_data:
        morans_i = replicate_data["morans_i"]
        morans_i_data.append(morans_i)
        replicate_names.append(replicate_data['dataset_name'])
    
    corr_matrix = pd.DataFrame(index=replicate_names, columns=replicate_names, dtype=float)
    min_common_probes = 20
    for i in range(len(morans_i_data)):
        for j in range(i + 1, len(morans_i_data)):
            common_probes = list(set(morans_i_data[i]["gene"]) & set(morans_i_data[j]["gene"]))
            print("In heatmap, number of common probes: ",len(common_probes))
            if len(common_probes) >= min_common_probes:
                #filter dataframes only for common probes then reorder them to have the same order
                morans_i_df = pd.merge(morans_i_data[i], morans_i_data[j], on="gene", suffixes=("_1", "_2"))
                morans_i_df = morans_i_df[morans_i_df["gene"].isin(common_probes)]
                corr = morans_i_df["Morans_I_1"].corr(morans_i_df["Morans_I_2"])
                corr=round(corr,3)
                corr_matrix.iloc[i, j] = corr
                corr_matrix.iloc[j, i] = corr
            else:
                corr_matrix.iloc[i, j] = 0.0
                corr_matrix.iloc[j, i] = 0.0

    #make diagonal 1
    np.fill_diagonal(corr_matrix.values, 1)
    print(corr_matrix)
    fig_dimension = len(replicates_data) 
    clustermap = sns.clustermap(corr_matrix, cmap='coolwarm', annot=True, fmt='.2f', figsize=(fig_dimension, fig_dimension))

    # Convert the plot to HTML
    img_buffer = BytesIO()
    clustermap.savefig(img_buffer, format='png')
    img_str = base64.b64encode(img_buffer.getvalue()).decode()
    html_fig = f'<img src="data:image/png;base64,{img_str}"/>'

    plt.close(clustermap.figure)
    return html_fig



def main():
    parser = argparse.ArgumentParser(description="Generate QC report for Pseudovisium output.")
    parser.add_argument("--folders", "-f", nargs="+", help="List of folders containing Pseudovisium output", required=True)
    parser.add_argument("--output_folder", "-o", default="/Users/k23030440/", help="Output folder path")
    parser.add_argument("--gene_names", "-g", nargs="+", default=["RYR3", "AQP4", "THBS1"], help="List of gene names to plot")
    parser.add_argument("--include_morans_i", "-m", action="store_true", help="Include Moran's I features tab")
    parser.add_argument("-max_workers", "--mw", type=int, default=4, help="Number of workers to use for parallel processing")
    args = parser.parse_args()

    generate_qc_report(args.folders, args.output_folder, args.gene_names, args.include_morans_i, max_workers=args.mw)


if __name__ == "__main__":
    main()



import csv
from collections import defaultdict
import numpy as np
import math
import pandas as pd
import cv2
import json
import gzip
import csv
import concurrent.futures
import os
import shutil
import tempfile
from tqdm import tqdm
import itertools
import psutil
import argparse
import tifffile
import imagecodecs
import multiprocessing
import time
import scanpy as sc
from scipy.sparse import csr_matrix
import concurrent.futures
import itertools
import csv
import multiprocessing
from tqdm import tqdm
from scipy.io import mmwrite


##Changes
#Save Fov information if provided
#Save quality score information if provided


def closest_hex(x,y,hexagon_size):
    """
    closest_hex(x, y, hexagon_size)
    Calculates the closest hexagon centroid to the given (x, y) coordinates.

    Args:
    x (float): The x-coordinate.
    y (float): The y-coordinate.
    hexagon_size (float): The size of the hexagon.

    Returns:
        tuple: The closest hexagon centroid coordinates (x, y) rounded to the nearest integer.
    """

    x_ = x // (hexagon_size*2)
    y_= y // (hexagon_size*1.732050807)


    if y_ % 2 == 1:

        #lower_left
        option_1_hexagon = (x_*2*hexagon_size,y_*1.732050807*hexagon_size)

        #lower right
        option_2_hexagon = ((x_+1)*2*hexagon_size,y_*1.732050807*hexagon_size)

        #upper middle
        option_3_hexagon = ((x_+0.5)*2*hexagon_size,(y_+1)*1.732050807*hexagon_size)

        #calc distance from each option and return the closest
        distance_1 = math.sqrt((x - option_1_hexagon[0])**2 + (y - option_1_hexagon[1])**2)
        distance_2 = math.sqrt((x - option_2_hexagon[0])**2 + (y - option_2_hexagon[1])**2)
        distance_3 = math.sqrt((x - option_3_hexagon[0])**2 + (y - option_3_hexagon[1])**2)
        closest = [option_1_hexagon,option_2_hexagon,option_3_hexagon][np.argmin([distance_1,distance_2,distance_3])]

    else:
            #lower middle 
        option_1_hexagon = ((x_+0.5)*2*hexagon_size,y_*(1.732050807*hexagon_size))

            #upper left
        option_2_hexagon = (x_*2*hexagon_size,(y_+1)*1.732050807*hexagon_size)

            #upper right
        option_3_hexagon = ((x_+1)*2*hexagon_size,(y_+1)*1.732050807*hexagon_size)

        #calc distance from each option and return the closest
        distance_1 = math.sqrt((x - option_1_hexagon[0])**2 + (y - option_1_hexagon[1])**2)
        distance_2 = math.sqrt((x - option_2_hexagon[0])**2 + (y - option_2_hexagon[1])**2)
        distance_3 = math.sqrt((x - option_3_hexagon[0])**2 + (y - option_3_hexagon[1])**2)
        closest = [option_1_hexagon,option_2_hexagon,option_3_hexagon][np.argmin([distance_1,distance_2,distance_3])]

    closest = (round(closest[0], 0), round(closest[1], 1))
    return closest



def initialize_hexagons(hexagon_size, field_size):
    """
    initialize_hexagons(hexagon_size, field_size)
    Initializes a dictionary of hexagon centroids for a given hexagon size and field size.

    Args:
    hexagon_size (float): The size of the hexagon.
    field_size (tuple): The size of the field as (field_size_x, field_size_y).

    Returns:
        dict: A dictionary where the keys are the hexagon centroid coordinates (x, y) and the values are the same coordinates.
    """



    print('initializing hexagons')
    # given an initial x and y coordinate, this function will create a zig-zag hexagonal lattice, where
    # hexagon size is the distance from the centroid to the vertex
    hexagons = {(0, 0): (0, 0)}
    # create hexagon centroids
    field_size_x, field_size_y = field_size
    for dy, i in zip(np.arange(-field_size_y * (1.732050807 * hexagon_size), (field_size_y) * (1.732050807 * hexagon_size), 1.732050807 * hexagon_size), [1, 2] * field_size_y):
        if i == 1:
            for dx in range((-field_size_x + 1) * hexagon_size, (field_size_x + 1) * hexagon_size, 2 * hexagon_size):

                hexagons[(dx, dy)] = (dx, dy)
        else:
            for dx in range((-field_size_x) * hexagon_size, (field_size_x) * hexagon_size, 2 * hexagon_size):
                
                hexagons[(dx, dy)] = (dx, dy)
    min_x = min([x for x, y in hexagons.keys()])
    min_y = min([y for x, y in hexagons.keys()])
    # rename hexagons both keys and values
    hexagons = {(round(x - min_x, 0), round(y - min_y, 1)): (round(x - min_x, 0), round(y - min_y, 1)) for (x, y) in hexagons.keys()}
    
    return hexagons


def preprocess_csv(csv_file, batch_size, fieldnames):
    """
    preprocess_csv(csv_file, batch_size, fieldnames)
    Preprocesses a CSV file by splitting it into smaller batches.

    Args:
        csv_file (str): The path to the CSV file.
        batch_size (int): The number of rows per batch.
        fieldnames (list): The list of field names to include in the batch CSV files.

    Returns:
        tuple: A tuple containing the temporary directory path and the total number of batches created.

    """


    tmp_dir = tempfile.mkdtemp(prefix="tmp_hexa")
    print(f"Created temporary directory {tmp_dir}")
    #if csv_file is .gz, then open it with gzip
    if csv_file.endswith('.gz'):
        with gzip.open(csv_file, 'rt') as file:
            reader = csv.DictReader(file)
            header = next(reader)  # Read the header row
            batch_num = 0

            while True:
                batch = list(itertools.islice(reader, batch_size))
                if not batch:
                    break  # Exit the loop if batch is empty

                batch_file = os.path.join(tmp_dir, f"batch_{batch_num}.csv")
                with open(batch_file, 'w', newline='') as batch_csv:
                    writer = csv.writer(batch_csv)
                    writer.writerow(fieldnames)  # Write header using the provided fieldnames
                    writer.writerows([[row[field] for field in fieldnames] for row in batch])  # Write rows

                batch_num += 1
                print(f"Created batch {batch_num}")

        print(f"Finished preprocessing. Total batches created: {batch_num}")
        return tmp_dir, batch_num


    with open(csv_file, 'r') as file:
        reader = csv.DictReader(file)
        header = next(reader)  # Read the header row
        batch_num = 0

        while True:
            batch = list(itertools.islice(reader, batch_size))
            if not batch:
                break  # Exit the loop if batch is empty

            batch_file = os.path.join(tmp_dir, f"batch_{batch_num}.csv")
            with open(batch_file, 'w', newline='') as batch_csv:
                writer = csv.writer(batch_csv)
                writer.writerow(fieldnames)  # Write header using the provided fieldnames
                writer.writerows([[row[field] for field in fieldnames] for row in batch])  # Write rows

            batch_num += 1
            print(f"Created batch {batch_num}")

    print(f"Finished preprocessing. Total batches created: {batch_num}")
    return tmp_dir, batch_num


def default_factory():

    return defaultdict(int)

def process_batch(batch_file, hexagon_size, feature_colname, x_colname, y_colname, cell_id_colname, quality_colname=None, quality_filter=False, count_colname="NA",smoothing=False, quality_per_hexagon=False, quality_per_probe=False):
    """
    process_batch(batch_file, hexagon_size, feature_colname, x_colname, y_colname, cell_id_colname, quality_colname=None, quality_filter=False, count_colname="NA", smoothing=False)
    Processes a batch CSV file to calculate hexagon counts and cell counts.

    Args:
    batch_file (str): The path to the batch CSV file.
    hexagon_size (float): The size of the hexagon.
    feature_colname (str): The name of the feature column.
    x_colname (str): The name of the x-coordinate column.
    y_colname (str): The name of the y-coordinate column.
    cell_id_colname (str): The name of the cell ID column.
    quality_colname (str, optional): The name of the quality score column. Defaults to None.
    quality_filter (bool, optional): Whether to filter rows based on quality score. Defaults to False.
    count_colname (str, optional): The name of the count column. Defaults to "NA".
    smoothing (bool, optional): Whether to apply smoothing to the counts. Defaults to False.

    Returns:
        dict or tuple: If cell_id_colname is not "None", returns a tuple containing the hexagon counts and hexagon cell counts.
                    Otherwise, returns the hexagon counts dictionary.

    """
    
    
    hexagon_counts = {}
    if cell_id_colname != "None":
        hexagon_cell_counts = {}
    if quality_per_hexagon==True:
        hexagon_quality = {}
    if quality_per_probe==True:
        probe_quality = {}
    if count_colname == "NA":
        print(f"Quality filter is set to {quality_filter}")
        print(f"Quality counting per hexagon is set to {quality_per_hexagon}")
        print(f"Quality counting per probe is set to {quality_per_probe}")

        ##########Xenium with quality filter scenario
        with open(batch_file, 'r') as file:
            reader = csv.DictReader(file)
            for row in reader:
                try:
                    x = float(row[x_colname])
                    y = float(row[y_colname])
                    closest_hexagon = closest_hex(x, y, hexagon_size)
                    if quality_per_hexagon==True:
                        if closest_hexagon not in hexagon_quality:
                            hexagon_quality[closest_hexagon] = {"mean":float(row[quality_colname]),"count":1}
                        else:
                            hexagon_quality[closest_hexagon]["mean"] = (hexagon_quality[closest_hexagon]["mean"]*hexagon_quality[closest_hexagon]["count"]+float(row[quality_colname]))/(hexagon_quality[closest_hexagon]["count"]+1)
                            hexagon_quality[closest_hexagon]["count"]+=1
                    if quality_per_probe==True:
                        if row[feature_colname] not in probe_quality:
                            probe_quality[row[feature_colname]] = {"mean":float(row[quality_colname]),"count":1}
                        else:
                            probe_quality[row[feature_colname]]["mean"] = (probe_quality[row[feature_colname]]["mean"]*probe_quality[row[feature_colname]]["count"]+float(row[quality_colname]))/(probe_quality[row[feature_colname]]["count"]+1)
                            probe_quality[row[feature_colname]]["count"]+=1

                    if quality_filter==True and float(row[quality_colname]) < 20:
                        continue
                    else:
                        if closest_hexagon not in hexagon_counts:
                            hexagon_counts[closest_hexagon] = {}
                        if row[feature_colname] not in hexagon_counts[closest_hexagon]:
                            hexagon_counts[closest_hexagon][row[feature_colname]] = 0
                        hexagon_counts[closest_hexagon][row[feature_colname]] += 1
                        if cell_id_colname != "None":
                            if closest_hexagon not in hexagon_cell_counts:
                                hexagon_cell_counts[closest_hexagon] = {}
                            if row[cell_id_colname] not in hexagon_cell_counts[closest_hexagon]:
                                hexagon_cell_counts[closest_hexagon][row[cell_id_colname]] = 0
                            hexagon_cell_counts[closest_hexagon][row[cell_id_colname]] += 1
                    

                except ValueError:
                    print(f"Skipping row due to invalid coordinates: {row}")
        
    else:
        if smoothing!=False:
            print("Smoothing is on. Counts will be spread into four neighboring squares")
            with open(batch_file, 'r') as file:
                ##########Visium HD with smoothing scenario
                reader = csv.DictReader(file)
                for row in reader:
                    try:
                        x = float(row[x_colname])
                        y = float(row[y_colname])
                        for x_new,y_new in [(x+smoothing,y+smoothing),(x-smoothing,y-smoothing),(x-smoothing,y+smoothing),(x+smoothing,y-smoothing)]:
                            closest_hexagon = closest_hex(x_new, y_new, hexagon_size)
                            if closest_hexagon not in hexagon_counts:
                                hexagon_counts[closest_hexagon] = {}
                            if row[feature_colname] not in hexagon_counts[closest_hexagon]:
                                hexagon_counts[closest_hexagon][row[feature_colname]] = 0
                            hexagon_counts[closest_hexagon][row[feature_colname]] += float(row[count_colname])/4
                            if cell_id_colname != "None":
                                if closest_hexagon not in hexagon_cell_counts:
                                    hexagon_cell_counts[closest_hexagon] = {}
                                if row[cell_id_colname] not in hexagon_cell_counts[closest_hexagon]:
                                    hexagon_cell_counts[closest_hexagon][row[cell_id_colname]] = 0
                                hexagon_cell_counts[closest_hexagon][row[cell_id_colname]] += float(row[count_colname])/4
                    except ValueError:
                        print(f"Skipping row due to invalid coordinates: {row}")
        else:
            print("No smoothing")
            with open(batch_file, 'r') as file:
                ##########Visium HD scenario without smoothing
                
                reader = csv.DictReader(file)
                for row in reader:
                    try:
                        x = float(row[x_colname])
                        y = float(row[y_colname])
                        closest_hexagon = closest_hex(x, y, hexagon_size)
                        if closest_hexagon not in hexagon_counts:
                            hexagon_counts[closest_hexagon] = {}
                        if row[feature_colname] not in hexagon_counts[closest_hexagon]:
                            hexagon_counts[closest_hexagon][row[feature_colname]] = 0
                        hexagon_counts[closest_hexagon][row[feature_colname]] += float(row[count_colname])
                        if cell_id_colname != "None":
                            if closest_hexagon not in hexagon_cell_counts:
                                hexagon_cell_counts[closest_hexagon] = {}
                            if row[cell_id_colname] not in hexagon_cell_counts[closest_hexagon]:
                                hexagon_cell_counts[closest_hexagon][row[cell_id_colname]] = 0
                            hexagon_cell_counts[closest_hexagon][row[cell_id_colname]] += float(row[count_colname])
                    except ValueError:
                        print(f"Skipping row due to invalid coordinates: {row}")

    returning_items = [hexagon_counts]

    if cell_id_colname != "None":
        returning_items.append(hexagon_cell_counts)
    if quality_per_hexagon==True:
        returning_items.append(hexagon_quality)
    if quality_per_probe==True:
        returning_items.append(probe_quality)
    return tuple(returning_items)



from concurrent.futures import ThreadPoolExecutor
from tqdm import tqdm
def process_batch_hexagons(batch,hexagon_counts,hexagon_names,features):
        batch_matrix_data = []
        for hexagon in batch:
            count_dict = hexagon_counts.get(hexagon, {})
            if sum(count_dict.values()) > 0:
                hexagon_index = hexagon_names.index(hexagon)
                for feature, count in count_dict.items():
                    feature_index = features.index(feature)
                    batch_matrix_data.append([feature_index+1, hexagon_index+1, count])
        return batch_matrix_data


import h5py
import numpy as np
from scipy.sparse import csr_matrix
from pathlib import Path
def write_10X_h5(adata, file):
    """Writes adata to a 10X-formatted h5 file.
    
    Note that this function is not fully tested and may not work for all cases.
    It will not write the following keys to the h5 file compared to 10X:
    '_all_tag_keys', 'pattern', 'read', 'sequence'

    Args:
        adata (AnnData object): AnnData object to be written.
        file (str): File name to be written to. If no extension is given, '.h5' is appended.

    Raises:
        FileExistsError: If file already exists.

    Returns:
        None
    """
    #remove file
    
    if '.h5' not in file: file = f'{file}.h5'
    if Path(file).exists():
        print(f"File `{file}` already exists. Removing it.")
        os.remove(file)
    def int_max(x):
        return int(max(np.floor(len(str(int(max(x)))) / 4), 1) * 4)
    def str_max(x):
        return max([len(i) for i in x])
    adata.var['feature_types'] = ['Gene Expression' for _ in range(adata.var.shape[0])]
    adata.var['genome'] = ['pv_placeholder' for _ in range(adata.var.shape[0])]
    adata.var['gene_ids'] = adata.var.index

    w = h5py.File(file, 'w')
    grp = w.create_group("matrix")
    grp.create_dataset("barcodes", data=adata.obs_names.values.astype('|S'))

    X = adata.X.T.tocsc()  # Convert the matrix to CSC format
    grp.create_dataset("data", data=X.data.astype(np.float32))
    grp.create_dataset("indices", data=X.indices.astype(np.int32))
    grp.create_dataset("indptr", data=X.indptr.astype(np.int32))
    grp.create_dataset("shape", data=np.array(X.shape).astype(np.int32))

    ftrs = grp.create_group("features")
    if 'feature_types' in adata.var:
        ftrs.create_dataset("feature_type", data=adata.var.feature_types.values.astype('|S'))
    if 'genome' in adata.var:
        ftrs.create_dataset("genome", data=adata.var.genome.values.astype('|S'))
    if 'gene_ids' in adata.var:
        ftrs.create_dataset("id", data=adata.var.gene_ids.values.astype('|S'))
    ftrs.create_dataset("name", data=adata.var.index.values.astype('|S'))
    
    w.close()

    


def process_csv_file(csv_file, hexagon_size, field_size, batch_size=1000000, technology="Xenium", feature_colname="feature_name", x_colname="x_location", y_colname="y_location", cell_id_colname="None", quality_colname="qv", max_workers=min(2, multiprocessing.cpu_count()),
                     quality_filter=False, count_colname="NA",smoothing=False,quality_per_hexagon=False,quality_per_probe=False,h5_x_colname="x",h5_y_colname="y"):
    """
    process_csv_file(csv_file, hexagon_size, field_size, batch_size=1000000, technology="Xenium", feature_colname="feature_name", x_colname="x_location", y_colname="y_location", cell_id_colname="None", quality_colname="qv", max_workers=min(2, multiprocessing.cpu_count()), quality_filter=False, count_colname="NA", smoothing=False)
    Processes a CSV file to calculate hexagon counts and cell counts using parallel processing.

    Args:
    csv_file (str): The path to the CSV file.
    hexagon_size (float): The size of the hexagon.
    field_size (tuple): The size of the field as (field_size_x, field_size_y).
    batch_size (int, optional): The number of rows per batch. Defaults to 1000000.
    technology (str, optional): The technology used. Defaults to "Xenium".
    feature_colname (str, optional): The name of the feature column. Defaults to "feature_name".
    x_colname (str, optional): The name of the x-coordinate column. Defaults to "x_location".
    y_colname (str, optional): The name of the y-coordinate column. Defaults to "y_location".
    cell_id_colname (str, optional): The name of the cell ID column. Defaults to "None".
    quality_colname (str, optional): The name of the quality score column. Defaults to "qv".
    max_workers (int, optional): The maximum number of worker processes to use. Defaults to min(2, multiprocessing.cpu_count()).
    quality_filter (bool, optional): Whether to filter rows based on quality score. Defaults to False.
    count_colname (str, optional): The name of the count column. Defaults to "NA".
    smoothing (bool, optional): Whether to apply smoothing to the counts. Defaults to False.

    Returns:
        tuple: A tuple containing the hexagon counts, hexagons dictionary, and hexagon cell counts.
    """





    hexagons = initialize_hexagons(hexagon_size, field_size)
    hexagon_counts = {hexagon: defaultdict(int) for hexagon in hexagons}
    hexagon_cell_counts = {hexagon: defaultdict(int) for hexagon in hexagons}
    hexagon_quality = {hexagon: defaultdict(int) for hexagon in hexagons}
    probe_quality = {}
    #create a nested dict called hexagon quality, with two keys for each hexagon, one for the quality score and one for the count
    
    if technology == "Xenium":
        print("Technology is Xenium. Going forward with default column names.")
        x_colname = "x_location"
        y_colname = "y_location"
        feature_colname = "feature_name"
        cell_id_colname = "cell_id"
        quality_colname = "qv"
        count_colname= "NA"

    elif technology == "Vizgen":
        print("Technology is Vizgen. Going forward with default column names.")
        x_colname = "global_x"
        y_colname = "global_y"
        feature_colname = "gene"
        #cell_id_colname = "barcode_id"
        count_colname= "NA"

    elif technology == "Nanostring":
        print("Technology is Nanostring. Going forward with default column names.")
        x_colname = "x_global_px"
        y_colname = "y_global_px"
        feature_colname = "target"
        cell_id_colname = "cell"
        count_colname= "NA"

    elif technology == "Visium_HD":
        print("Technology is Visium_HD. Going forward with pseudovisium processed colnames.")
        x_colname = "x"
        y_colname = "y"
        feature_colname = "gene"
        #cell_id_colname = "barcode"
        count_colname= "count"

    elif technology == "seqFISH":
        print("Technology is seqFISH. Going forward with default column names.")
        x_colname = "x"
        y_colname = "y"
        feature_colname = "name"
        cell_id_colname = "cell"
        count_colname= "NA"
    
    elif technology == "Curio":
        print("Technology is Curio. Going forward with default column names.")
        x_colname = "x"
        y_colname = "y"
        feature_colname = "gene"
        cell_id_colname = "barcode"
        count_colname= "count"


    else:
        print("Technology not recognized. Going forward with set column names.")

    fieldnames=[feature_colname, x_colname, y_colname]
    if cell_id_colname!="None":
        fieldnames.append(cell_id_colname)
    if quality_filter==True or quality_per_hexagon==True or quality_per_probe==True:
        fieldnames.append(quality_colname)
    if count_colname!="NA":
        fieldnames.append(count_colname)



    tmp_dir, num_batches = preprocess_csv(csv_file, batch_size, fieldnames)
    batch_files = [os.path.join(tmp_dir, f"batch_{i}.csv") for i in range(num_batches)]
    with concurrent.futures.ProcessPoolExecutor(max_workers=min(max_workers, multiprocessing.cpu_count())) as executor:
        futures = [executor.submit(process_batch, batch_file, hexagon_size, feature_colname, x_colname, y_colname, cell_id_colname, quality_colname, quality_filter, count_colname,smoothing,quality_per_hexagon,quality_per_probe) for batch_file in batch_files]
        
        with tqdm(total=len(batch_files), desc="Processing batches", unit="batch") as progress_bar:
            for future in concurrent.futures.as_completed(futures):
                all_res = future.result()
                batch_hexagon_counts = all_res[0]
                for hexagon_counts_hex, counts in batch_hexagon_counts.items():
                    for feature_name, count in counts.items():
                        try:
                            hexagon_counts[hexagon_counts_hex][feature_name] += count
                        except KeyError:
                            print(f"Feature {feature_name} not found in hexagon_counts")

                if cell_id_colname != "None":
                    batch_hexagon_cell_counts = all_res[1]
                    for hexagon_cell_counts_hex, cell_counts in batch_hexagon_cell_counts.items():
                        for cell_id, cell_count in cell_counts.items():
                            try:
                                hexagon_cell_counts[hexagon_cell_counts_hex][cell_id] += cell_count
                            except KeyError:
                                print(f"Cell ID {cell_id} not found in hexagon_cell_counts")

                if quality_per_hexagon==True:
                    if cell_id_colname != "None":
                        batch_hexagon_quality = all_res[2]
                    else:
                        batch_hexagon_quality = all_res[1]
                    for hexagon_quality_hex, quality_dict in batch_hexagon_quality.items():
                        try:
                            if hexagon_quality_hex not in hexagon_quality:
                                hexagon_quality[hexagon_quality_hex] = quality_dict
                            else:
                                hexagon_quality[hexagon_quality_hex]["mean"] = (float(hexagon_quality[hexagon_quality_hex]["mean"]) * hexagon_quality[hexagon_quality_hex]["count"] + float(quality_dict["mean"]) * quality_dict["count"]) / (hexagon_quality[hexagon_quality_hex]["count"] + quality_dict["count"])
                                hexagon_quality[hexagon_quality_hex]["count"] += quality_dict["count"]
                        except KeyError:
                            print(f"Quality not found in hexagon_quality for hexagon {hexagon_quality_hex}")
                
                if quality_per_probe==True:
                    if cell_id_colname != "None" and quality_per_hexagon==True:
                        batch_probe_quality = all_res[3]
                    elif cell_id_colname != "None" and quality_per_hexagon==False:
                        batch_probe_quality = all_res[2]
                    elif cell_id_colname == "None" and quality_per_hexagon==True:
                        batch_probe_quality = all_res[2]
                    else:
                        batch_probe_quality = all_res[1]
                

                    for probe, quality_dict in batch_probe_quality.items():
                        try:
                            if probe not in probe_quality:
                                probe_quality[probe] = quality_dict
                            else:
                                probe_quality[probe]["mean"] = (float(probe_quality[probe]["mean"]) * probe_quality[probe]["count"] + float(quality_dict["mean"]) * quality_dict["count"]) / (probe_quality[probe]["count"] + quality_dict["count"])
                                probe_quality[probe]["count"] += quality_dict["count"]
                        except KeyError:
                            print(f"Quality not found in probe_quality for probe {probe}")
                progress_bar.update(1)

    shutil.rmtree(tmp_dir)  # Remove temporary directory and files

    return hexagon_counts, hexagons, hexagon_cell_counts, hexagon_quality, probe_quality

from concurrent.futures import ThreadPoolExecutor
from tqdm import tqdm

def hex_to_rows(hexagon_batch, hexagon_indices, features, hexagon_counts):
    """
    hex_to_rows(hexagon_batch, hexagon_indices, features, hexagon_counts)
    Converts a batch of hexagons to matrix rows.

    Args:
        hexagon_batch (list): A list of hexagon coordinates.
        hexagon_indices (list): A list of hexagon indices.
        features (list): A list of feature names.
        hexagon_counts (dict): A dictionary of hexagon counts.

    Returns:
        list: A list of matrix rows, where each row contains the feature index, hexagon index, and count.
    """
    def process_hexagon(hexagon, hexagon_index):
        return [[features.index(feature) + 1, hexagon_index + 1, count]
                for feature, count in hexagon_counts[hexagon].items()]

    matrix_data = [item for hexagon, hexagon_index in zip(hexagon_batch, hexagon_indices)
                   for item in process_hexagon(hexagon, hexagon_index)]
    return matrix_data

def create_pseudovisium(path,hexagon_counts,hexagon_cell_counts,hexagon_quality, probe_quality,
                        img_file_path=None,  project_name="project",
                         alignment_matrix_file=None,image_pixels_per_um=1/0.2125,hexagon_size=100,tissue_hires_scalef=0.2,
                         pixel_to_micron=False,max_workers=min(2, multiprocessing.cpu_count())):
    """
    create_pseudovisium(path, hexagon_counts, hexagon_cell_counts, img_file_path=None, project_name="project", alignment_matrix_file=None, image_pixels_per_um=1/0.2125, hexagon_size=100, tissue_hires_scalef=0.2, pixel_to_micron=False, max_workers=min(2, multiprocessing.cpu_count()))
    Creates a Pseudovisium output directory structure and files.
    
    Args:
    path (str): The path to create the Pseudovisium output directory.
    hexagon_counts (dict): A dictionary of hexagon counts.
    hexagon_cell_counts (dict): A dictionary of hexagon cell counts.
    img_file_path (str, optional): The path to the image file. Defaults to None.
    project_name (str, optional): The name of the project. Defaults to "project".
    alignment_matrix_file (str, optional): The path to the alignment matrix file. Defaults to None.
    image_pixels_per_um (float, optional): The number of image pixels per micrometer. Defaults to 1/0.2125.
    hexagon_size (int, optional): The size of the hexagon. Defaults to 100.
    tissue_hires_scalef (float, optional): The scaling factor for the high-resolution tissue image. Defaults to 0.2.
    pixel_to_micron (bool, optional): Whether to convert pixel coordinates to micron coordinates. Defaults to False.
    max_workers (int, optional): The maximum number of worker processes to use. Defaults to min(2, multiprocessing.cpu_count()).
    """

    #to path, create a folder called pseudovisium
    folderpath = path+ '/pseudovisium/' + project_name
    #if folderpath exists, delete it
    if os.path.exists(folderpath):
        shutil.rmtree(folderpath)
    try:
        print("Creating pseudovisium folder in output path:{0}".format(folderpath))
 
        if os.path.exists(path+ '/pseudovisium/'):
            print("Using already existing folder: {0}".format(path+ '/pseudovisium/'))
        else:
            os.mkdir(path+ '/pseudovisium/')
        os.mkdir(folderpath)
        os.mkdir(folderpath + '/spatial')
    except:
        pass
    
 ############################################## ##############################################

    # see https://kb.10xgenomics.com/hc/en-us/articles/11636252598925-What-are-the-Xenium-image-scale-factors
    #https://www.10xgenomics.com/support/software/space-ranger/latest/analysis/outputs/spatial-outputs
    scalefactors = {"tissue_hires_scalef":tissue_hires_scalef,
                     "tissue_lowres_scalef": tissue_hires_scalef/10,
                       "fiducial_diameter_fullres": 0,
                         "spot_diameter_fullres": 2*hexagon_size*(image_pixels_per_um)}
    
    print("Creating scalefactors_json.json file in spatial folder.")
    with open(folderpath +'/spatial/scalefactors_json.json', 'w') as f:
        json.dump(scalefactors, f)
 ############################################## ##############################################

    x, y, x_, y_, contain, hexagon_names = [], [], [], [], [], []
    for hexagon in hexagon_counts:
        x_.append((hexagon[0] + hexagon_size) // (2 * hexagon_size))
        y_.append(hexagon[1] // (1.73205 * hexagon_size))
        x.append(hexagon[0])
        y.append(hexagon[1])
        contain.append(1 if sum(hexagon_counts[hexagon].values()) > hexagon_size else 0)
        hexagon_names.append(hexagon)
    

 ############################################## ##############################################
    barcodes = ["hexagon_{}".format(i) for i in range(1, len(hexagon_names) + 1)]
    barcodes_table = pd.DataFrame({'barcode':barcodes})
    #save to pseudo visium root
    barcodes_table.to_csv(folderpath+'/barcodes.tsv',sep='\t',index=False,header=False)

    print("Creating barcodes.tsv.gz file in spatial folder.")
    with open(folderpath +'/barcodes.tsv', 'rb') as f_in:
        with gzip.open(folderpath +'/barcodes.tsv.gz', 'wb') as f_out:
            f_out.writelines(f_in)
 ############################################## ##############################################


    hexagon_table = pd.DataFrame(zip(barcodes, contain, y_, x_, 
                                 [int(image_pixels_per_um * a) for a in y],
                                 [int(image_pixels_per_um * a) for a in x]),
                             columns=['barcode', 'in_tissue', 'array_row', 'array_col',
                                      'pxl_row_in_fullres', 'pxl_col_in_fullres'])
    
    print("Creating tissue_positions_list.csv file in spatial folder.")
    hexagon_table.to_csv(folderpath +'/spatial/tissue_positions_list.csv',index=False,header=False)

 ############################################## ##############################################
    #if hexagon_cell_counts is empty, then skip
    if hexagon_cell_counts == {}:
        print("No cell information provided. Skipping cell information files.")
    else:
        print("Creating pv_cell_hex.csv file in spatial folder.")
        with open(folderpath + '/spatial/pv_cell_hex.csv', 'w', newline='') as f:
            writer = csv.writer(f)
            for hexagon, cell_count_dict in hexagon_cell_counts.items():
                if sum(cell_count_dict.values()) > 0:
                    hexagon_index = hexagon_names.index(hexagon)
                    for cell, count in cell_count_dict.items():
                        writer.writerow([cell, hexagon_index + 1, count])

 ############################################## ##############################################
    if hexagon_quality == {}:
        print("No quality information provided. Skipping quality information files.")
    else:
        print("Creating quality_per_hexagon.csv file in spatial folder.")
        with open(folderpath + '/spatial/quality_per_hexagon.csv', 'w', newline='') as f:
            writer = csv.writer(f)
            for hexagon, quality_dict in hexagon_quality.items():
                hexagon_index = hexagon_names.index(hexagon)
                writer.writerow([hexagon_index + 1, quality_dict["mean"], quality_dict["count"]])

 ############################################## ##############################################
    if probe_quality == {}:
        print("No quality information provided. Skipping quality information files.")
    else:
        print("Creating quality_per_probe.csv file in spatial folder.")
        with open(folderpath + '/spatial/quality_per_probe.csv', 'w', newline='') as f:
            writer = csv.writer(f)
            for probe, quality_dict in probe_quality.items():
                writer.writerow([probe, quality_dict["mean"], quality_dict["count"]])




 ############################################## ##############################################

    #get all unique feature_names
    features = list(set(feature for hexagon_counts in hexagon_counts.values() for feature in hexagon_counts))

    # Create a list of rows with repeated features and 'Gene Expression' column
    rows = [[feature, feature, 'Gene Expression'] for feature in features]

    print("Creating features.tsv.gz file in spatial folder.")
    with open(folderpath + '/features.tsv', 'wt', newline='', encoding='utf-8') as f_out:
        writer = csv.writer(f_out, delimiter='\t')
        writer.writerows(rows)
    
    # Create a features.tsv.gz file
    with open(folderpath + '/features.tsv', 'rb') as f_in, gzip.open(folderpath + '/features.tsv.gz', 'wb') as f_out:
        f_out.writelines(f_in)
        

 ############################################## ##############################################
    

    print("Putting together the matrix.mtx file")
    matrix_data = []
    batch_size_n_hexagons = 500
    hexagon_names = list(hexagon_counts.keys())  # Convert hexagon_names to a list
    hexagon_indices = list(range(len(hexagon_names)))  # Create an array of hexagon indices
    n_total_hexagons = len(hexagon_names)
    n_batches = n_total_hexagons // batch_size_n_hexagons
    n_leftover = n_total_hexagons % batch_size_n_hexagons

    # Create batches of hexagon names and indices
    hexagon_batches = [hexagon_names[i*batch_size_n_hexagons:(i+1)*batch_size_n_hexagons] for i in range(n_batches)]
    index_batches = [hexagon_indices[i*batch_size_n_hexagons:(i+1)*batch_size_n_hexagons] for i in range(n_batches)]
    if n_leftover > 0:
        hexagon_batches.append(hexagon_names[-n_leftover:])
        index_batches.append(hexagon_indices[-n_leftover:])

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = []
        for hexagon_batch, index_batch in zip(hexagon_batches, index_batches):
            future = executor.submit(hex_to_rows, hexagon_batch, index_batch, features, hexagon_counts)
            futures.append(future)

        with tqdm(total=len(hexagon_batches), desc="Processing batches") as pbar:
            for future in futures:
                matrix_data.extend(future.result())
                pbar.update(1)
    

    data = np.array(matrix_data)[:, 2]
    row_indices = np.array(matrix_data)[:, 0] - 1
    col_indices = np.array(matrix_data)[:, 1] - 1

    sparse_matrix = csr_matrix((data, (row_indices, col_indices)), shape=(len(features), len(barcodes)))

    print("Creating matrix.mtx.gz file in spatial folder.")
    with open(folderpath + '/matrix.mtx', 'wb') as f:
        mmwrite(f, sparse_matrix, comment='metadata_json: {"software_version": "Pseudovisium", "format_version": 1}\n')

    with open(folderpath +'/matrix.mtx', 'rb') as f_in, gzip.open(folderpath + '/matrix.mtx.gz', 'wb') as f_out:
        f_out.writelines(f_in)
    
 ############################################## ##############################################



    print("Putting together the filtered_feature_bc_matrix.h5 file")
    data = np.array(matrix_data)[:, 2]
    row_indices = np.array(matrix_data)[:, 1] - 1  # Use hexagon indices for rows
    col_indices = np.array(matrix_data)[:, 0] - 1  # Use feature indices for columns
    sparse_matrix = csr_matrix((data, (row_indices, col_indices)), shape=(len(barcodes), len(features)))

    # Create AnnData object from sparse matrix and barcodes/features
    adata = sc.AnnData(X=sparse_matrix, obs=pd.DataFrame(index=barcodes), var=pd.DataFrame(index=features))

    # Write AnnData object to 10X-formatted h5 file
    write_10X_h5(adata, folderpath + '/filtered_feature_bc_matrix.h5')

############################################## ##############################################

    #check if alignment matrix file is given
    if alignment_matrix_file:
        print("Alignment matrix found and will be used to create tissue_hires_image.png and tissue_lowres_image.png files in spatial folder.")
        M = pd.read_csv(alignment_matrix_file,header=None,index_col=None).to_numpy()
    else:
        M = np.array([[1.0,0.0,0.0],[0.0,1.0,0.0],[0.0,0.0,1.0]])
            # Load the H&E image
    if img_file_path:
        #if img_filepath is tiff
        if img_file_path.endswith('.tiff') or img_file_path.endswith('.tif'):
            image = tifffile.imread(img_file_path)
        #elif png
        elif img_file_path.endswith('.png'):
            image = cv2.imread(img_file_path, cv2.IMREAD_UNCHANGED)
        #resizing the image according to tissue_hires_scalef, but such that it also satisfies the incoming
        #scaling factor from the alignment matrix

        if pixel_to_micron:
            M[0, 0] = 1/M[0, 0]   # Update x-scale to 1 because rescaling already done here
            M[1, 1] = 1/M[1, 1]   # Update y-scale to 1 because rescaling already done here
            width = int(image.shape[1] * tissue_hires_scalef) #* M[0, 0])  # New width
            height = int(image.shape[0] * tissue_hires_scalef) #* M[0, 0])  # New height
            dim = (width, height)
        # resize image
            resized_img = cv2.resize(image, dim, interpolation = cv2.INTER_AREA)

            #update translation
            M[0, 2] *= tissue_hires_scalef * M[0, 1]
            M[1, 2] *= tissue_hires_scalef * M[0, 1]

            M[0, 2] = -M[0, 2]
            M[1, 2] = -M[1, 2]

        else: #normal scenario

            width = int(image.shape[1] * tissue_hires_scalef )  # New width
            height = int(image.shape[0] * tissue_hires_scalef)  # New height
            dim = (width, height)
            # resize image
            resized_img = cv2.resize(image, dim, interpolation = cv2.INTER_AREA)

            #update translation
            M[0, 2] *= tissue_hires_scalef #* max(M[0, 0], M[1, 1])  # Update x-translation
            M[1, 2] *= tissue_hires_scalef #* max(M[0, 0], M[1, 1])  # Update y-translation

        max_dim = max(resized_img.shape)
        max_stretch = max(M[0, 0], M[1, 1],M[1, 0],M[0, 1] )*1.2
                # Apply the transformation
        new_width = int(max_dim*max_stretch)
        new_height = int(max_dim*max_stretch)
        image = cv2.warpAffine(resized_img, M[:2], (new_width, new_height))


        #and to 2%
        scale2 = 0.1
        width = int(image.shape[1] * scale2)
        height = int(image.shape[0] * scale2)
        dim = (width, height)
        resized = cv2.resize(image, dim, interpolation = cv2.INTER_AREA)

        #finally if image is a single channel, convert to three channels, because some packages will expect that
        #when importing.
        dims = len(image.shape)
        #if 2, then triple the image to make it thre channels
        if dims == 2:
            image = np.array([image,image,image])
            #change order of axis
            image = np.moveaxis(image,0,-1)

            resized = np.array([resized,resized,resized])
            #change order of axis
            resized = np.moveaxis(resized,0,-1)


        print("Creating tissue_hires_image.png file in spatial folder.")
        #save as 8bit
        cv2.imwrite(folderpath +'/spatial/tissue_hires_image.png', image/np.max(image)*255)
        cv2.imwrite(folderpath +'/spatial/tissue_lowres_image.png', resized/np.max(resized)*255)
    
    else:
        #output a blank image
        print("No image file provided. Creating blank tissue_hires_image.png and tissue_lowres_image.png files in spatial folder.")
        image = np.zeros((1000,1000))
        cv2.imwrite(folderpath +'/spatial/tissue_hires_image.png', image)
        cv2.imwrite(folderpath +'/spatial/tissue_lowres_image.png', image)
        


#####################
###############  Visium HD/Slide-seq helper functions 
###################################################################################################################################################
def read_files(folder,technology):
    """
    read_files(folder)
    Reads the necessary files from a Visium HD folder.

    Args:
    folder (str): The path to the Visium HD folder.

    Returns:
        tuple: A tuple containing the scale factors, tissue positions, and filtered feature-barcode matrix.
    """
    if technology == "Visium_HD":
        scalefactors = json.load(open(folder+"/spatial/scalefactors_json.json"))
        tissue_pos = pd.read_parquet(folder+"/spatial/tissue_positions.parquet")
        fb_matrix = sc.read_10x_h5(folder+"/filtered_feature_bc_matrix.h5")
        return scalefactors,tissue_pos,fb_matrix
    elif technology == "Curio":
        #find file ending with .h5ad in folder
        folder_files = os.listdir(folder)
        h5ad_file = [file for file in folder_files if file.endswith(".h5ad")][0]
        adata = sc.read(folder+h5ad_file)
        return adata
    

def anndata_to_df(adata, technology,tissue_pos=None,scalefactors=None,x_col=None,y_col=None):
    """
    anndata_to_df(adata, tissue_pos, scalefactors)
    Converts an AnnData object to a DataFrame.
    Args:
    adata (AnnData): The AnnData object containing the counts matrix.
    tissue_pos (DataFrame): The tissue positions DataFrame.
    scalefactors (dict): The scale factors dictionary.

    Returns:
        tuple: A tuple containing the converted DataFrame and the image resolution.
    """
    if technology == "Visium_HD":
        #get image resolution of the hires image
        image_resolution = scalefactors["microns_per_pixel"]/scalefactors["tissue_hires_scalef"]
        #change to pixel per micron
        image_resolution = 1/image_resolution
        #tissue_pos keep barcode, pxl_row_in_fullres, pxl_col_in_fullres
        tissue_pos = tissue_pos[["barcode","pxl_row_in_fullres","pxl_col_in_fullres"]]
        # Convert the AnnData matrix to a sparse matrix (CSR format)
        X = csr_matrix(adata.X)

        # Get the row and column indices of non-zero elements
        row_indices, col_indices = X.nonzero()

        # Create a DataFrame with the row names, column names, and counts
        df = pd.DataFrame({
            'barcode': adata.obs_names[row_indices],
            'gene': adata.var_names[col_indices],
            'count': X.data
        })

        #join with tissue_pos
        df = pd.merge(df,tissue_pos,on="barcode")

        #extract microns_per_pixel
        microns_per_pixel = scalefactors["microns_per_pixel"]
        #convert pxl_row_in_fullres and pxl_col_in_fullres to microns
        df["y"] = df["pxl_row_in_fullres"]*microns_per_pixel
        df["x"] = df["pxl_col_in_fullres"]*microns_per_pixel
        #remove pxl_row_in_fullres and pxl_col_in_fullres
        df.drop(["pxl_row_in_fullres","pxl_col_in_fullres"],axis=1,inplace=True)

        # Write the DataFrame to a CSV file
        return df,image_resolution 
    elif technology == "Curio":
        #get scales
        
        #keep only the columns x,y
        obs = adata.obs[[x_col,y_col]]
        #rename these cols to x and y
        obs.columns = ["x","y"]
        scale = np.round(min([abs(x1 - x2) for x1 in obs["y"] for x2 in [obs["y"][500]] if x1!=x2]),3) #should be 10um
        
        obs["barcode"] = obs.index
        X = csr_matrix(adata.X)

        # Get the row and column indices of non-zero elements
        row_indices, col_indices = X.nonzero()
        genes = adata.var.index if (adata.var.index[0] != 0) and (adata.var.index[0] != "0") else adata.var[adata.var.columns[0]]
        genes = genes[col_indices]
        # Create a DataFrame with the row names, column names, and counts
        df = pd.DataFrame({
            'barcode': adata.obs_names[row_indices],
            'gene': genes,
            'count': X.data
        })
        #join with tissue_pos
        df = pd.merge(df,obs,on="barcode")
        return df,scale


def visium_hd_curio_to_transcripts(folder,output,technology,x_col=None,y_col=None):
    """
    visium_hd_to_transcripts(folder, output)
    Converts Visium HD files to a transcripts CSV file.
    Args:
    folder (str): The path to the Visium HD folder.
    output (str): The path to save the transcripts CSV file.

    Returns:
        float: The image resolution (pixels per micrometer).
    """

    if technology == "Visium_HD":
        scalefactors,tissue_pos,fb_matrix = read_files(folder)
        df,image_resolution  = anndata_to_df(fb_matrix,tissue_pos,scalefactors)
        df.to_csv(output,index=False)
        return image_resolution 
    elif technology == "Curio":
        adata = read_files(folder,technology)
        df,scale = anndata_to_df(adata,technology,x_col=x_col,y_col=y_col)
        df.to_csv(output,index=False)
        return scale





######### Main function to generate pseudovisium output ############################################################################################################
    
def generate_pv(csv_file,img_file_path=None, hexagon_size=100, field_size_x=1000, 
                field_size_y=1000, output_path=None, batch_size=1000000, alignment_matrix_file=None, project_name='project',
                image_pixels_per_um=1/0.85, tissue_hires_scalef=0.2,technology="Xenium", 
                feature_colname="feature_name", x_colname="x_location", y_colname="y_location",
                cell_id_colname="None", quality_colname="qv",
                pixel_to_micron=False, max_workers=min(2, multiprocessing.cpu_count()), quality_filter=False, count_colname="NA",visium_hd_folder=None,
                smoothing=False,quality_per_hexagon=False,quality_per_probe=False,h5_x_colname = "x", h5_y_colname = "y"):
    """
    generate_pv(csv_file, img_file_path=None, hexagon_size=100, field_size_x=1000, field_size_y=1000, output_path=None, 
    batch_size=1000000, alignment_matrix_file=None, project_name='project', image_pixels_per_um=1/0.85, tissue_hires_scalef=0.2, 
    technology="Xenium", feature_colname="feature_name", x_colname="x_location", y_colname="y_location", cell_id_colname="None", 
    quality_colname="qv", pixel_to_micron=False, max_workers=min(2, multiprocessing.cpu_count()), quality_filter=False, 
    count_colname="NA", visium_hd_folder=None, smoothing=False)

    
    Generates a Pseudovisium output from a CSV file.

    Args:
    csv_file (str): The path to the CSV file.
    img_file_path (str, optional): The path to the image file. Defaults to None.
    hexagon_size (int, optional): The size of the hexagon. Defaults to 100.
    field_size_x (int, optional): The size of the field in the x-dimension. Defaults to 1000.
    field_size_y (int, optional): The size of the field in the y-dimension. Defaults to 1000.
    output_path (str, optional): The path to save the Pseudovisium output. Defaults to None.
    batch_size (int, optional): The number of rows per batch. Defaults to 1000000.
    alignment_matrix_file (str, optional): The path to the alignment matrix file. Defaults to None.
    project_name (str, optional): The name of the project. Defaults to 'project'.
    image_pixels_per_um (float, optional): The number of image pixels per micrometer. Defaults to 1/0.85.
    tissue_hires_scalef (float, optional): The scaling factor for the high-resolution tissue image. Defaults to 0.2.
    technology (str, optional): The technology used. Defaults to "Xenium".
    feature_colname (str, optional): The name of the feature column. Defaults to "feature_name".
    x_colname (str, optional): The name of the x-coordinate column. Defaults to "x_location".
    y_colname (str, optional): The name of the y-coordinate column. Defaults to "y_location".
    cell_id_colname (str, optional): The name of the cell ID column. Defaults to "None".
    quality_colname (str, optional): The name of the quality score column. Defaults to "qv".
    pixel_to_micron (bool, optional): Whether to convert pixel coordinates to micron coordinates. Defaults to False.
    max_workers (int, optional): The maximum number of worker processes to use. Defaults to min(2, multiprocessing.cpu_count()).
    quality_filter (bool, optional): Whether to filter rows based on quality score. Defaults to False.
    count_colname (str, optional): The name of the count column. Defaults to "NA".
    visium_hd_folder (str, optional): The path to the Visium HD folder. Defaults to None.
    smoothing (float, optional): The smoothing factor. Defaults to False.
    """


    start = time.time()

    if technology == "Visium_HD":
        print("Technology is Visium_HD. Generating transcripts.csv file from Visium HD files.")
        #Automatically calculating image_pixels_per_um from the scalefactors_json.json file
        image_pixels_per_um = visium_hd_curio_to_transcripts(visium_hd_folder,visium_hd_folder+"/transcripts.csv",technology)
        csv_file = visium_hd_folder+"/transcripts.csv"
    if technology == "Curio":
        print("Technology is Curio. Generating transcripts.csv file from Curio files.")
        smoothing_scale = visium_hd_curio_to_transcripts(visium_hd_folder,visium_hd_folder+"/transcripts.csv",technology,x_col=h5_x_colname,y_col=h5_y_colname)
        csv_file = visium_hd_folder+"/transcripts.csv"
        print("Smoothing defaults to : {0}".format(smoothing_scale/4))
        smoothing = smoothing_scale/4
        
        # Process CSV file to generate hexagon counts and hexagon information
    field_size = (field_size_x, field_size_y)
    hexagon_counts, hexagons, hexagon_cell_counts, hexagon_quality, probe_quality= process_csv_file(csv_file, hexagon_size, field_size, batch_size, 
             technology, feature_colname, x_colname, y_colname,cell_id_colname, quality_colname=quality_colname,
               max_workers=max_workers, quality_filter=quality_filter, count_colname=count_colname,smoothing=smoothing,
               quality_per_hexagon=quality_per_hexagon,quality_per_probe=quality_per_probe,h5_x_colname=h5_x_colname,h5_y_colname=h5_y_colname)
        
    # Create Pseudovisium output
    create_pseudovisium(path=output_path,hexagon_counts=hexagon_counts,hexagon_cell_counts=hexagon_cell_counts, probe_quality=probe_quality,
                        img_file_path=img_file_path, hexagon_quality =hexagon_quality , 
                          project_name=project_name, alignment_matrix_file=alignment_matrix_file,
                          image_pixels_per_um=image_pixels_per_um,hexagon_size=hexagon_size,
                          tissue_hires_scalef=tissue_hires_scalef,pixel_to_micron=pixel_to_micron,max_workers=max_workers)

    #save all arguments in a json file called arguments.json
    print("Creating arguments.json file in output path.")
    arguments = {"csv_file":csv_file,"img_file_path":img_file_path,"hexagon_size":hexagon_size,
                    "field_size_x":field_size_x,"field_size_y":field_size_y,"output_path":output_path,
                    "batch_size":batch_size,"alignment_matrix_file":alignment_matrix_file,"project_name":project_name,
                    "image_pixels_per_um":image_pixels_per_um,"tissue_hires_scalef":tissue_hires_scalef,
                    "technology":technology,"feature_colname":feature_colname,"x_colname":x_colname,
                    "y_colname":y_colname,"cell_id_colname":cell_id_colname,"pixel_to_micron":pixel_to_micron,
                    "quality_colname":quality_colname,"quality_filter":quality_filter,"count_colname":count_colname,
                    "smoothing":smoothing,"quality_per_hexagon":quality_per_hexagon,"quality_per_probe":quality_per_probe,
                    "max_workers":max_workers,"visium_hd_folder":visium_hd_folder,"h5_x_colname":h5_x_colname,"h5_y_colname":h5_y_colname}

    with open(output_path + '/pseudovisium/' + project_name + '/arguments.json', 'w') as f:
        json.dump(arguments, f)

    end = time.time()
    print(f"Time taken: {end - start} seconds")




def main():
    """
    main()
    The main function that parses command-line arguments and calls the generate_pv function.
    """
    parser = argparse.ArgumentParser(description="Process parameters.")
    parser.add_argument("--csv_file", "-c", type=str, help="CSV file path", default=None)
    parser.add_argument("--output_path", "-o", type=str, help="Output path", default='.')
    parser.add_argument("--hexagon_size", "-hs", type=int, help="Hexagon size", default=100)
    parser.add_argument("--field_size_x", "-fsx", type=int, help="Field size x", default=1000)
    parser.add_argument("--field_size_y", "-fsy", type=int, help="Field size y", default=1000)
    parser.add_argument("--img_file_path", "-i", type=str, help="Image file path", default=None)
    parser.add_argument("--alignment_matrix_file", "-am", type=str, help="Alignment matrix file path", default=None)
    parser.add_argument("--batch_size", "-b", type=int, help="Batch size", default=1000000)
    parser.add_argument("--project_name", "-p", type=str, help="Project name", default='project')
    parser.add_argument("--image_pixels_per_um", "-ppu", type=float, help="Image pixels per um", default=1/0.2125)#change!
    parser.add_argument("--tissue_hires_scalef", "-ths", type=float, help="Tissue hires scale factor", default=0.2)
    parser.add_argument("--technology", "-t", type=str, help="Technology", default="Xenium")
    parser.add_argument("--feature_colname", "-fc", type=str, help="Feature column name", default="feature_name")
    parser.add_argument("--x_colname", "-xc", type=str, help="X column name", default="x_location")
    parser.add_argument("--y_colname", "-yc", type=str, help="Y column name", default="y_location")
    parser.add_argument("--cell_id_colname", "-cc", type=str, help="Cell ID column name", default="None")
    parser.add_argument("--pixel_to_micron", "-ptm", action="store_true", help="Convert pixel to micron")
    parser.add_argument("--quality_colname", "-qcol", type=str, help="Quality column name", default="qv")
    parser.add_argument("--count_colname", "-ccol", type=str, help="Count column name", default="NA")
    parser.add_argument("--smoothing", "-s", type=float, help="Smoothing factor", default=0.0)
    parser.add_argument("--visium_hd_folder", "-vhf", type=str, help="Visium HD folder", default=None)
    parser.add_argument("--mw", "--max_workers", type=int, help="Max workers", default=min(2, multiprocessing.cpu_count()))
    parser.add_argument("--quality_filter", "-qf", action="store_true", help="Filter out rows with quality score less than 20")
    parser.add_argument("--quality_per_hexagon", "-qph", action="store_true", help="Calculate quality per hexagon")
    parser.add_argument("--quality_per_probe", "-qpp", action="store_true", help="Calculate quality per probe")
    parser.add_argument("--h5_x_colname", "-h5x", type=str, help="X column name in h5ad file", default="x")
    parser.add_argument("--h5_y_colname", "-h5y", type=str, help="Y column name in h5ad file", default="y")
    parser.add_argument("-v", "--verbose", action="store_true", help="Print out script purpose and parameters")
    
    
    #make sure to add verbose as well
    parser.add_argument("-help", action="store_true", help="Print out script purpose and parameters")
    args = parser.parse_args()

    if args.help:
        print("This is Pseudovisium, a software that compresses imaging-based spatial transcriptomics files using hexagonal binning of the data.")
        parser.print_help()
        sys.exit(0)

    #print value of quality filter
    print(f"Quality filter is set to {args.quality_filter}")
    

    generate_pv(csv_file=args.csv_file,img_file_path=args.img_file_path,
                hexagon_size=args.hexagon_size, field_size_x=args.field_size_x, 
                field_size_y=args.field_size_y, output_path=args.output_path, 
                batch_size=args.batch_size,
                alignment_matrix_file=args.alignment_matrix_file, 
                project_name=args.project_name,image_pixels_per_um=args.image_pixels_per_um, 
                tissue_hires_scalef=args.tissue_hires_scalef,technology=args.technology, 
                feature_colname=args.feature_colname, x_colname=args.x_colname, 
                y_colname=args.y_colname,cell_id_colname=args.cell_id_colname,
                pixel_to_micron=args.pixel_to_micron,max_workers=args.mw,
                quality_colname=args.quality_colname,quality_filter=args.quality_filter,
                count_colname=args.count_colname,
                smoothing=args.smoothing,
                quality_per_hexagon=args.quality_per_hexagon,
                quality_per_probe=args.quality_per_probe,
                h5_x_colname=args.h5_x_colname,h5_y_colname=args.h5_y_colname,
                visium_hd_folder=args.visium_hd_folder)
                
    print("Pseudovisium output generated successfully.")


    


if __name__ == "__main__":
    main()




#this is a setup.py file
from setuptools import setup, find_packages


VERSION = '0.0.3'
DESCRIPTION = 'Python package for hexagonal binning of high-resolution spatial transcriptomic data'
LONG_DESCRIPTION = 'Python package for hexagonal binning of high-resolution spatial transcriptomic data, for more information see https://github.com/BKover99/pseudovisium'


setup(
    name="Pseudovisium",
    version=VERSION,
    author="Bence Kover",
    author_email="<kover.bence@gmail.com>",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=['numpy','pandas', 'scanpy', 'squidpy', 'matplotlib'],
    keywords=['spatial', 'transcriptomics', 'visium', 'xenium', 'cosmx'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X"
    ]
)

